"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import time
import logging
import warnings
import pyomo.environ as pyomo
from pyomo.solvers.plugins.solvers.persistent_solver import PersistentSolver
from pyomo.opt import SolverStatus, TerminationCondition

from pycity_scheduling.classes import CityDistrict, Building, ElectricalGrid
from pycity_scheduling.exceptions import NonoptimalError
from pycity_scheduling.solvers import DEFAULT_SOLVER, DEFAULT_SOLVER_OPTIONS


class OptimizationAlgorithm:
    """
    Base class for all optimization algorithms.

    This class provides functionality common to all algorithms which are
    able to optimize City Districts.

    Parameters
    ----------
    city_district : CityDistrict
    solver : str, optional
        Solver to use for solving (sub-)problems.
    solver_options : dict, optional
        Options to pass to calls to the solver. Keys are the name of
        the functions being called and are one of `__call__`, `set_instance`,
        `solve`.

        - `__call__` is the function being called when generating an instance
          with the pyomo SolverFactory. In addition to the options provided,
          `node_ids` is passed to this call containing the IDs of the entities
          being optimized.
        - `set_instance` is called when a pyomo Model is set as an instance of
          a persistent solver.
        - `solve` is called to perform an optimization. If not set,
          `save_results` and `load_solutions` may be set to false to provide a
          speedup.

    mode : str, optional
        Specifies which set of constraints to use.

        - `convex`  : Use linear, i.e., convex, constraints
        - `integer` : May use non-linear, i.e., mixed-integer, constraints
    """

    def __init__(self, city_district, solver=DEFAULT_SOLVER, solver_options=DEFAULT_SOLVER_OPTIONS, mode="convex"):
        self.city_district = city_district
        self.entities = [city_district]
        self.entities.extend([node["entity"] for node in city_district.nodes.values()])
        if self.city_district.electrical_grid is not None:
            self.entities.append(self.city_district.electrical_grid)
        self.solver = solver
        self.solver_options = solver_options
        self.mode = mode
        self._reset()

    def _add_objective(self):
        """Adds the modified objective of the entities to their specific models."""
        raise NotImplementedError("This method should be implemented by subclass.")

    def solve(self, full_update=True, beta=1, robustness=None, debug=False):
        """Solves the city district for the current op_horizon.

        Parameters
        ----------
        full_update : bool, optional
            Should be true if the city district models were changed or
            update_model should be called to update the city district models.
            Disabling the full_update can give a small performance gain.
        beta : float, optional
            Tradeoff factor between system and customer objective. The customer
            objective is multiplied with beta.
        robustness : tuple, optional
            Tuple of two floats. First entry defines how many time steps are
            protected from deviations. Second entry defines the magnitude of
            deviations which are considered.
        debug : bool, optional
            Specify whether detailed debug information shall be printed.

        Returns
        -------
        results : dict
            Dictionary of performance values of the algorithm.

        Raises
        ------
        NonoptimalError
            If no feasible solution for the city district is found or a solver
            problem is encountered.
        """
        results, params = self._presolve(full_update, beta, robustness, debug)
        params["start_time"] = time.monotonic()
        self._solve(results, params, debug)
        self._postsolve(results, params, debug)
        return results

    def _save_time(self, results, params):
        """Saves the current runtime into results."""
        results["times"].append(time.monotonic() - params["start_time"])
        return

    @staticmethod
    def _get_beta(params, entity):
        """Returns the beta value for a specific entity."""
        beta = params["beta"]
        if isinstance(beta, dict):
            return beta.get(entity.id, 1.0)
        if isinstance(entity, CityDistrict):
            return 1.0
        return beta

    def _reset(self):
        """Reset, i.e., 'unpopulate', the previous optimization model."""
        city_district = self.entities[0]
        city_district.model = None
        for bd in city_district.get_lower_entities():
            bd.model = None
            if isinstance(bd, Building) and bd.has_bes:
                bd.bes.model = None
                for e in bd.bes.get_lower_entities():
                    e.model = None
            if isinstance(bd, Building) and len(bd.apartments) == 1:
                bd.apartments[0].model = None
                for e in bd.apartments[0].get_lower_entities():
                    e.model = None
            elif isinstance(bd, Building) and len(bd.apartments) > 1:
                for ap in bd.apartments:
                    ap.model = None
                    for e in ap.get_lower_entities():
                        e.model = None

    def _presolve(self, full_update, beta, robustness, debug):
        """Step before the optimization of (sub-)problems.

        Parameters
        ----------
        full_update : bool
            Should be true if the city district models were changed or
            update_model should be called to update the city district models.
            Disabling the full_update can give a small performance gain.
        beta : float
            Tradeoff factor between system and customer objective. The customer
            objective is multiplied with beta.
        robustness : tuple, optional
            Tuple of two floats. First entry defines how many time steps are
            protected from deviations. Second entry defines the magnitude of
            deviations which are considered.
        debug : bool, optional
            Specify whether detailed debug information shall be printed.

        Returns
        -------
        results : dict
            Dictionary in which performance values of the algorithm can be stored.
        params : dict
            Dictionary in which the algorithm can store intermediate results for
            later access in the algorithm itself. This dictionary should contain
            all which is generated and used by the algorithm.
        """
        params = {"beta": beta, "robustness": robustness}
        results = {"times": []}
        return results, params

    def _solve(self, results, params, debug):
        """Step in which (sub-)problems are optimized.

        Parameters
        ----------
        results : dict
            Dictionary in which performance values of the algorithm are stored.
        params : dict
            Dictionary in which intermediate results are stored.
        debug : bool
            Specify whether detailed debug information shall be printed.
        """
        raise NotImplementedError("This method should be implemented by the subclass.")

    def _postsolve(self, results, params, debug):
        """Step after optimization.

        In this step other post-processing can be done.

        Parameters
        ----------
        results : dict
            Dictionary in which performance values of the algorithm are stored.
        params : dict
            Dictionary in which intermediate results are stored.
        debug : bool
            Specify whether detailed debug information shall be printed.
        """
        return


class IterationAlgorithm(OptimizationAlgorithm):
    """Base class for all optimization algorithms that solve the problem iteratively."""
    def _solve(self, results, params, debug):
        if "iterations" not in results:
            results["iterations"] = []
        if "obj_value" not in results:
            results["obj_value"] = []
        iterations = results["iterations"]
        is_last_iteration = False
        while not is_last_iteration:
            iterations.append((iterations[-1] + 1) if len(iterations) > 0 else 1)
            self._iteration(results, params, debug=debug)
            self._save_time(results, params)
            is_last_iteration = self._is_last_iteration(results, params, debug)
        return

    def _iteration(self, results, params, debug):
        """Execute a single iteration of the algorithm.

        Parameters
        ----------
        results : dict
            Dictionary in which performance values of the algorithm are stored.
        params : dict
            Dictionary in which intermediate results are stored.
        """
        return

    def _is_last_iteration(self, results, params, debug):
        """Returns True if the current iteration is the last one.
            It checks if the iteration limit is exceeded.
            Overwrite this function to apply additional, individual stopping criteria other than the iteration limit.
            Anyway, this parent method must be called in a child method, so that the check of an exceeded iteration
            limit below is executed.

        Parameters
        ----------
        debug : bool, optional
            Specify whether detailed debug information shall be printed.
        Raises
        ------
        Warning
            If the stopping criteria can not be reached in max_iterations.
        NonoptimalError
            If no feasible solution for the city district is found or a solver
            problem is encountered.
        """
        if results["iterations"][-1] >= self.max_iterations:
            if debug:
                warnings.warn("User defined iteration limit exceeded")
            print("Exceeded the user defined iteration limit of {0} iterations. "
                  "Terminating the iterative algorithm.".format(self.max_iterations))
            return True
        return False


class DistributedAlgorithm(OptimizationAlgorithm):
    """Base class for all distributed optimization algorithms.

    These algorithms can divide the optimization problem into sub-problems.
    """
    def _solve_nodes(self, results, params, nodes, variables=None, debug=True):
        """Used to indicate which nodes can be solved independently.

        Provides the "distributed_times" as a performance value to results.

        Parameters
        ----------
        results : dict
            Dictionary in which performance values of the algorithm are stored.
        params : dict
            Dictionary in which intermediate results are stored.
        nodes : list of SolverNode
            List of nodes which can be solved independently.
        variables : list of lists of variables, optional
            Can contain a list for each node in nodes to indicate to pyomo which
            variables should be loaded back into the model. Specifying this can
            lead to a significant speedup.
        debug : bool, optional
            Specify whether detailed debug information shall be printed. Defaults
            to true.
        """
        if "distributed_times" not in results:
            results["distributed_times"] = []
        if variables is None:
            variables = [None] * len(nodes)
        node_times = {}
        for node, variables_ in zip(nodes, variables):

            for entity in node.entities:
                if (str(entity.id)+"_times") not in results:
                    results[str(entity.id)+"_times"] = []

            start = time.monotonic()
            node.solve(debug=debug)
            stop = time.monotonic()

            entity_ids = tuple(entity.id for entity in node.entities)
            node_times[entity_ids] = stop - start
            for entity in node.entities:
                results[str(entity.id)+"_times"].append(stop - start)
        results["distributed_times"].append(node_times)
        return


class SolverNode:
    """Node which can be used to solve all entities provided to it.

    Provides an abstraction layer for algorithms, so entities can be
    assigned to nodes and optimized easily.

    Parameters
    ----------
    solver : str
        Solver to use for solving (sub)problems.
    solver_options : dict, optional
        Options to pass to calls to the solver. Keys are the name of
        the functions being called and are one of `__call__`, `set_instance_`,
        `solve`.

        - `__call__` is the function being called when generating an instance
          with the pyomo SolverFactory.  Additionally to the options provided,
          `node_ids` is passed to this call containing the IDs of the entities
          being optimized.
        - `set_instance` is called when a pyomo Model is set as an instance of
          a persistent solver.
        - `solve` is called to perform an optimization. If not set,
          `save_results` and `load_solutions` may be set to false to provide a
          speedup.
    entities : list
        List of entities which should be optimized by this node.
    mode : str, optional
        Specifies which set of constraints to use.

        - `convex`  : Use linear constraints
        - `integer`  : May use non-linear constraints
    robustness : tuple, optional
        Tuple of two floats. First entry defines how many time steps are
        protected from deviations. Second entry defines the magnitude of
        deviations which are considered.
    """
    def __init__(self, solver, solver_options, entities, mode="convex", robustness=None):
        logging.getLogger('pyomo.core').setLevel(logging.ERROR)
        self.solver = pyomo.SolverFactory(solver, node_ids=[entity.id for entity in entities],
                                          **solver_options.get("__call__", {}))
        self.solver_options = solver_options
        self.is_persistent = isinstance(self.solver, PersistentSolver)
        self.robustness = robustness
        self.entities = entities
        self.mode = mode
        self.model = None
        self._prepare()

    def _prepare(self):
        """Create the pyomo model for the entities and populate it."""
        model = pyomo.ConcreteModel()
        for entity in self.entities:
            self._prepare_model(entity, model, robustness=self.robustness)
        self.model = model
        return

    def _prepare_model(self, entity, model, robustness=None):
        """Add a single entity to a model."""
        if isinstance(entity, Building):
            entity.populate_model(model, self.mode, robustness=robustness)
            entity.update_model(self.mode, robustness=robustness)
        else:
            entity.populate_model(model, self.mode)
            entity.update_model(self.mode)
        return

    def full_update(self, robustness=None):
        """Execute the update_model function and propagate other model changes.

        Parameters
        ----------
        robustness : tuple, optional
            Tuple of two floats. First entry defines how many time steps are
            protected from deviations. Second entry defines the magnitude of
            deviations which are considered.
        """
        for entity in self.entities:
            if isinstance(entity, Building):
                entity.update_model(mode=self.mode, robustness=robustness)
            else:
                entity.update_model(mode=self.mode)
        if self.is_persistent:
            self.solver.set_instance(self.model, **self.solver_options.get("set_instance", {}))
        return

    def constr_update(self):
        """Only propagate the constraints update of the model."""
        if self.is_persistent:
            self.solver.set_instance(self.model, **self.solver_options.get("set_instance", {}))
        else:
            pass
        return

    def obj_update(self):
        """Only propagate the objective value update of the model."""
        if self.is_persistent:
            self.solver.set_objective(self.model.o)
        else:
            pass
        return

    def solve(self, debug=True):
        """Call the solver to solve this nodes optimization problem.

        Parameters
        ----------
        debug : bool, optional
            Specify whether detailed debug information shall be printed. Defaults
            to true.
        """
        solve_options = self.solver_options.get("solve", {})
        if self.is_persistent:
            result = self.solver.solve(**solve_options)
        else:
            result = self.solver.solve(self.model, **solve_options)
        if (not (result.solver.termination_condition == TerminationCondition.optimal or
                result.solver.termination_condition == TerminationCondition.maxTimeLimit or
                result.solver.termination_condition == TerminationCondition.maxIterations) or
                not (result.solver.status == SolverStatus.ok or
                     result.solver.status == SolverStatus.warning or
                     result.solver.status == SolverStatus.aborted)):
            if debug:
                import pycity_scheduling.util.debug as debug
                debug.analyze_model(self.model, self.solver, result)
            raise NonoptimalError("Could not retrieve schedule from model.")
        for entity in self.entities:
            if not isinstance(entity, ElectricalGrid):
                entity.update_schedule()
        return



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import pyomo.environ as pyomo

import ray
import copy
import psutil
import time

from pycity_scheduling.classes import (CityDistrict, Building, Photovoltaic, WindEnergyConverter)
from pycity_scheduling.algorithms.algorithm import IterationAlgorithm, DistributedAlgorithm, SolverNode
from pycity_scheduling.solvers import DEFAULT_SOLVER, DEFAULT_SOLVER_OPTIONS
from pycity_scheduling.algorithms.algorithm_ray import RayDerivativeFreeALADINSolverNode, list_into_n_chunks


class DerivativeFreeALADINRay(IterationAlgorithm, DistributedAlgorithm):
    """
    Implementation of the distributed  ALADIN algorithm using ray.io parallelization.
    Specifically, this class implements the derivative-free ALADIN as described in [1].

    Parameters
    ----------
    city_district : CityDistrict
    solver : str, optional
        Solver to use for solving (sub)problems.
    solver_options : dict, optional
        Options to pass to calls to the solver. Keys are the name of
        the functions being called and are one of `__call__`, `set_instance_`,
        `solve`.
        `__call__` is the function being called when generating an instance
        with the pyomo SolverFactory. In addition to the options provided,
        `node_ids` is passed to this call containing the IDs of the entities
        being optimized.
        `set_instance` is called when a pyomo Model is set as an instance of
        a persistent solver. `solve` is called to perform an optimization. If
        not set, `save_results` and `load_solutions` may be set to false to
        provide a speedup.
    mode : str, optional
        Specifies which set of constraints to use.
        - `convex`  : Use linear constraints
        - `integer`  : May use non-linear constraints
    eps_primal : float, optional
        Primal stopping criterion for the ALADIN algorithm.
    eps_dual : float, optional
        Dual stopping criterion for the ALADIN algorithm.
    rho : float, optional
        Penalty term parameter for the ALADIN algorithm.
    alpha : float, optional
        Step size parameter for the ALADIN algorithm.
    max_iterations : int, optional
        Maximum number of ALADIN iterations.
    robustness : tuple, optional
        Tuple of two floats. First entry defines how many time steps are
        protected from deviations. Second entry defines the magnitude of
        deviations which are considered.
    ray_cpu_count : int, optional
        Number of CPU cores to be used by ray for parallelization.
        Default: Detect the number of CPUs automatically.

    References
    ----------
    [1] "Distributed Optimization and Control with ALADIN" by Boris Houska and Yuning Jiang
    Online: https://faculty.sist.shanghaitech.edu.cn/faculty/boris/paper/AladinChapter.pdf (accessed on 2024/02/13)
    """
    def __init__(self, city_district, solver=DEFAULT_SOLVER, solver_options=DEFAULT_SOLVER_OPTIONS, mode="convex",
                 eps_primal=0.001, eps_dual=0.001, rho=0.5, alpha=1.0, hessian_scaling=1.0, max_iterations=10000,
                 robustness=None, ray_cpu_count=None):
        super(DerivativeFreeALADINRay, self).__init__(city_district, solver, solver_options, mode)
        self.eps_primal = eps_primal
        self.eps_dual = eps_dual
        self.hessian_scaling = hessian_scaling
        self.rho = rho
        self.alpha = alpha
        self.max_iterations = max_iterations
        self.op_horizon = self.city_district.op_horizon

        # Only consider entities of type CityDistrict, Building, Photovoltaic, WindEnergyConverter
        self._entities = [entity for entity in self.entities if
                          isinstance(entity, (CityDistrict, Building, Photovoltaic, WindEnergyConverter))]

        # Initialize ray before using ray functions
        self.ray_cpu_count = ray_cpu_count

        # Ray initialization on a single machine vs. on an active ray cluster:
        try:
            ray.shutdown()
            if self.ray_cpu_count is not None:
                ray.init(ignore_reinit_error=True, num_cpus=int(self.ray_cpu_count))
            else:
                ray.init(ignore_reinit_error=True, num_cpus=min(len(self._entities), psutil.cpu_count(logical=False)))
        except:
            ray.init(ignore_reinit_error=True)
        resources = ray.cluster_resources()

        # Create a remote ray actor solver node for each available CPU core
        node_num = len(ray.nodes())
        if self.ray_cpu_count is not None:
            cpu_count = self.ray_cpu_count
        else:
            cpu_count = int(max(resources["CPU"], psutil.cpu_count(logical=False)))
        entity_chunks = copy.deepcopy(list(list_into_n_chunks(self._entities,
                                                              int(min(cpu_count, len(self._entities))))))
        print("{} nodes available | {} CPUs available/defined | {} entities are present | {} entity chunks created".
            format(node_num, cpu_count, len(self._entities), len(entity_chunks)))

        self.nodes = []
        self.global_entity_indices = dict()

        for i, chunk in zip(range(len(entity_chunks)), entity_chunks):
            for chunk_entity in chunk:
                # Find index of entity in self._entities and safe it to the dictionary
                global_entity_index = (
                    next((i for i, item in enumerate(self._entities) if item.id == chunk_entity.id), -1))
                self.global_entity_indices.update({chunk_entity.id: global_entity_index})
            self.nodes.append(RayDerivativeFreeALADINSolverNode.remote(solver, solver_options,
                                                                       entities=chunk,
                                                                       rho=self.rho,
                                                                       hessian_scaling=self.hessian_scaling,
                                                                       node_index=i,
                                                                       op_horizon=self.op_horizon,
                                                                       mode=mode,
                                                                       robustness=robustness))

        # Add objectives on all remotes
        add_obj_refs = [node.add_objective.remote() for node in self.nodes]

        # Wait until pyomo parameter calculation is finished on all nodes
        while len(add_obj_refs) > 0:
            done_add_obj_refs, add_obj_refs = ray.wait(add_obj_refs)
            for ref in done_add_obj_refs:
                ray.get(ref)

        # Initialize the ALADIN QP model:
        self.qp_solver = pyomo.SolverFactory(solver, **solver_options.get("__call__", {}))
        self.qp_solver_options = solver_options
        self.qp_model = pyomo.ConcreteModel()
        self.qp_model.size_t = pyomo.RangeSet(0, self.op_horizon - 1)
        self.qp_model.size_n = pyomo.RangeSet(0, len(self._entities) - 1)
        self.qp_model.size_n_t = pyomo.Set(initialize=self.qp_model.size_n * self.qp_model.size_t)
        self.qp_model.size_n_t_t = pyomo.Set(
            initialize=self.qp_model.size_n * self.qp_model.size_t * self.qp_model.size_t)
        self.qp_model.x = pyomo.Param(self.qp_model.size_n_t, mutable=True, initialize=0)
        self.qp_model.delta_x = pyomo.Var(self.qp_model.size_n_t, domain=pyomo.Reals, initialize=0)
        self.qp_model.gradient_k = pyomo.Param(self.qp_model.size_n_t, mutable=True, initialize=0)
        self.qp_model.hessian_k = pyomo.Param(self.qp_model.size_n_t_t, mutable=True, initialize=0)

        self._add_constraint_qp()
        self._add_objective_qp()

    def _add_objective_qp(self):
        obj = 0.0
        for i in range(len(self._entities)):
            # add Hessian part to QP objective
            for j in range(self.op_horizon):
                for k in range(self.op_horizon):
                    obj += self.qp_model.delta_x[i, j] * self.qp_model.hessian_k[i, j, k] * self.qp_model.delta_x[i, k]

            # add gradient part to QP objective
            for j in range(self.op_horizon):
                obj += self.qp_model.gradient_k[i, j] * self.qp_model.delta_x[i, j]
        self.qp_model.o = pyomo.Objective(expr=obj)
        self.qp_model.dual = pyomo.Suffix(direction=pyomo.Suffix.IMPORT)
        return

    def _add_constraint_qp(self):
        def qp_constraint(model, t):
            sum = 0.0
            for i in range(len(self._entities)):
                sum += model.x[i, t] + model.delta_x[i, t]
            return sum == 0.0
        self.qp_model.qp_constr = pyomo.Constraint(self.qp_model.size_t, rule=qp_constraint)
        return

    def _presolve(self, full_update, beta, robustness, debug):
        results, params = super()._presolve(full_update, beta, robustness, debug)

        set_model_beta_refs = []
        # Set the model betas on each node
        for i, node in enumerate(self.nodes):
            set_model_beta_refs.append(node.set_model_betas.remote(params))
            if full_update:
                set_model_beta_refs.append(node.full_update.remote(robustness))

        while len(set_model_beta_refs):
            done_set_model_beta_refs, set_model_beta_refs = ray.wait(set_model_beta_refs)
            for ref in done_set_model_beta_refs:
                ray.get(ref)

        results["QP_times"] = []
        
        results["r_norms"] = []
        results["s_norms"] = []

        for entity_id in self.global_entity_indices.keys():
            results[str(self.global_entity_indices[entity_id]) + "_times"] = []

        return results, params

    def _postsolve(self, results, params, debug):
        # ------------------------------------------
        # 1) Remote post solve
        # ------------------------------------------
        # Execute a remote version of the postsolve method on each node,
        # the local schedules of each entity are the calculated
        ray.get([node.postsolve.remote() for node in self.nodes])

        schedule_dict = dict()
        get_schedules_refs = [node.get_all_schedules.remote() for node in self.nodes]
        # ------------------------------------------
        # 2) Wait for remote post solves
        # ------------------------------------------
        # Wait until all nodes are finished and retrieve there local dictionary
        # The dictionary contains each asset id with its corresponding schedules
        while len(get_schedules_refs) > 0:
            done_get_schedules_refs, get_schedules_refs = ray.wait(get_schedules_refs)
            for done_schedule in done_get_schedules_refs:
                node_schedules = ray.get(done_schedule)
                schedule_dict.update(node_schedules)

        # ------------------------------------------
        # 3) Local schedule update
        # ------------------------------------------
        # Update the local entities with the schedules calculated on the remote nodes
        for entity in self._entities:
            if not isinstance(entity, CityDistrict):
                for asset in entity.get_all_entities():
                    if hasattr(asset, "id"):
                        schedules = schedule_dict.get(asset.id)
                        if schedules is not None:
                            asset.schedules = schedules
            if hasattr(entity, "id"):
                schedules = schedule_dict.get(entity.id)
                if schedules is not None:
                    entity.schedules = schedules

        city_district = self._entities[0]
        for entity in city_district.get_all_entities():
            if hasattr(entity, "id"):
                schedules = schedule_dict.get(entity.id)
                if schedules is not None:
                    entity.schedules = schedules

        # Terminate all ray actors
        for node in self.nodes:
            node.exit.remote()

        if ray.is_initialized is True:
            ray.shutdown()

        # inversion of the sign of the final aggregator schedule:
        self.city_district.p_el_schedule = -self.city_district.p_el_schedule
        return

    def _is_last_iteration(self, results, params, debug):
        if super(DerivativeFreeALADINRay, self)._is_last_iteration(results, params, debug):
            return True
        return results["r_norms"][-1] <= self.eps_primal and results["s_norms"][-1] <= self.eps_dual

    def _iteration(self, results, params, debug):
        super(DerivativeFreeALADINRay, self)._iteration(results, params, debug)

        # fill parameters if not already present
        if "x" not in params:
            params["x"] = np.zeros((len(self._entities), self.op_horizon))
        if "delta_x" not in params:
            params["delta_x"] = np.zeros((len(self._entities), self.op_horizon))
        if "lambda_k" not in params:
            params["lambda_k"] = np.zeros(self.op_horizon)
        if "x_k" not in params:
            params["x_k"] = np.zeros((len(self._entities), self.op_horizon))
        if "gradient_k" not in params:
            params["gradient_k"] = np.zeros((len(self._entities), self.op_horizon))
        if "hessian_k" not in params:
            params["hessian_k"] = np.zeros((len(self._entities), self.op_horizon, self.op_horizon))

        # ------------------------------------------
        # 0) Storage for variables
        # ------------------------------------------
        p_el_schedules = np.zeros((len(self._entities), self.op_horizon))
        lambda_k = params["lambda_k"]
        x_k = params["x_k"]

        lambda_k_stored = ray.put(lambda_k)
        x_k_stored = ray.put(x_k)
        global_entity_indices_stored = ray.put(self.global_entity_indices)

        # ------------------------------------------
        # 1) Optimize all entities
        # ------------------------------------------
        solve_ids = [node.solve.remote(variable_refs=[lambda_k_stored, x_k_stored, global_entity_indices_stored],
                                       debug=debug) for i, node in zip(range(len(self.nodes)), self.nodes)]

        obj_value = 0
        while len(solve_ids):
            done_solve_ids, solve_ids = ray.wait(solve_ids)
            for done_solve_id in done_solve_ids:
                returns = ray.get(done_solve_id)
                obj_value += returns[2]
                for entity_id, entity_value, sub_time_val in zip(returns[0], returns[1], returns[3]):
                    global_index = self.global_entity_indices[entity_id]
                    p_el_schedules[global_index] = entity_value
                    results[str(global_index)+"_times"].append(sub_time_val)

        # extract solutions ("p_el" values) for each sub-problem
        for i in range(len(self._entities)):
            params["x"][i] = p_el_schedules[i]

        r = sum(p_el_schedules[:])
        s = np.zeros_like(params["x"])
        for i in range(len(self._entities)):
            s[i] = (params["x"][i] - params["x_k"][i])
        results["r_norms"].append(np.linalg.norm(r, 2))
        results["s_norms"].append(np.linalg.norm(s.flatten(), 2))

        # ------------------------------------------
        # 2) Solve centralized QP
        # ------------------------------------------
        t_0 = time.monotonic()
        for i in range(len(self._entities)):
            params["hessian_k"][i] = self.hessian_scaling * np.eye(self.op_horizon)  # ToDo: make it adjustable
            params["gradient_k"][i] = np.matmul(params["hessian_k"][i],
                                                params["x_k"][i] - params["x"][i]) - params["lambda_k"]

            for j in range(self.op_horizon):
                self.qp_model.x[i, j] = params["x"][i][j]
                self.qp_model.gradient_k[i, j] = params["gradient_k"][i][j]
                for k in range(self.op_horizon):
                    self.qp_model.hessian_k[i, j, k] = params["hessian_k"][i][j][k]

        if self.solver == "gurobi_direct" or self.solver == "gurobi":
            qp_solver_options = self.qp_solver_options.get("solve", {})
            self.qp_solver.solve(self.qp_model, **qp_solver_options)
        elif self.solver == "gurobi_persistent":
            self.qp_solver.set_instance(self.qp_model)
            qp_solver_options = self.qp_solver_options.get("solve", {})
            self.qp_solver.solve(**qp_solver_options)
        else:
            raise "Gurobi solver supported only!"
        
        t_1 = time.monotonic()
        results['QP_times'].append(t_1 - t_0)

        # extract solutions ("delta_x" and "lambda_k" values) for the centralized QP
        dual_qp = -np.array([self.qp_model.dual[self.qp_model.qp_constr[t]] for t in range(self.op_horizon)])
        for i in range(len(self._entities)):
            for t in range(self.op_horizon):
                params["delta_x"][i][t] = pyomo.value(self.qp_model.delta_x[i, t])

        # ------------------------------------------
        # 3) Update of ALADIN variables
        # ------------------------------------------
        params["lambda_k"] = params["lambda_k"] + self.alpha * (dual_qp - params["lambda_k"])

        for i in range(len(self._entities)):
            for t in range(self.op_horizon):
                params["x_k"][i][t] = params["x"][i][t] + self.alpha * params["delta_x"][i][t]

        results["obj_value"].append(obj_value)
        return


"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import pyomo.environ as pyomo
from warnings import warn

from shapely import bounds

from pycity_scheduling.classes.battery import Battery
from pycity_scheduling import util
from pycity_scheduling import constants


class ElectricVehicle(Battery):
    """
    Class representing an electric vehicle for scheduling purposes.

    Parameters
    ----------
    environment : Environment
        Common Environment instance.
    e_el_max : float
        Electric capacity of the battery in [kWh].
    p_el_max_charge : float
        Maximum charging power in [kW].
    p_el_max_discharge : float, optional
        Maximum discharging power in [kW]. Defaults to zero.
    soc_init : float, optional
        Initial state of charge. Defaults to 50%.
    charging_time : array of binaries, optional
        Indicator when electric vehicle can be charged.

        - `charging_time[t] == 0`: EV cannot be charged in t
        - `charging_time[t] == 1`: EV can be charged in t

        It must contain at least one `0` otherwise the model will become
        infeasible. Its length has to be consistent with `ct_pattern`.
        Defaults to only charge during night.
    ct_pattern : str, optional
        Define how the `charging_time` profile is to be used.

        - `None` : Profile matches simulation horizon (default).
        - 'daily' : Profile matches one day.
        - 'weekly' : Profile matches one week.
    simulate_driving : bool, optional
        Simulate driving.

        - `True`: EV may drive and lose energy during the trip
        - `False`: EV cannot drive and lose energy during the trip
    minimum_soc_end : float, optional
        Set the minimum state of charge for an EV at the end of the optimization horizon. Must be in (0,1].
        Defaults to 80%.
    eta : float, optional
        Charging and discharging efficiency of the EV. Must be in (0,1].
        Defaults to 100%.

    Notes
    -----
    - EVs offer sets of constraints for operation. The :math:`e_{el}` equivalence
      constraint is replaced by the following constraint:

    .. math::
        e_{el} &=& e_{el\\_previous} + (\\eta * p_{el\\_demand}
        - (1 / \\eta) * p_{el\\_supply} - p_{el\\_drive}) * \\Delta t \\\\

    - The following constraints are added:

    .. math::
        p_{el\\_drive} \\geq 0 \\\\
        p_{el\\_demand\\_i} = p_{el\\_supply} = 0,
        & \\quad \\text{if} \\quad ct\\_pattern_i = 0 \\\\
        p_{el\\_drive\\_i} = 0, & \\quad \\text{if} \\quad ct\\_pattern_i = 1 \\\\
        e_{el\\_i} = soc\\_init * e_{el\\_max}, & \\quad \\text{if} \\quad \\sum_{j=0}^i ct\\_pattern_j = 0 \\\\
        e_{el\\_i} = 0.2 * e_{el\\_max}, & \\quad \\text{else if} \\quad ct\\_pattern_i = 0 \\\\
        e_{el\\_i} = e_{el\\_max}, & \\quad \\text{else if} \\quad ct\\_pattern_i+1 = 0

    - The constraint for the parameter `storage_end_equality` is removed. Instead,
      the EV needs to be fully charged at the end of the `simu_horizon` if parameter `ct_pattern`
      is one at the end of the simulation horizon.
    """

    def __init__(self, environment, e_el_max, p_el_max_charge, p_el_max_discharge=0.0,
                 soc_init=0.5, charging_time=None, ct_pattern=None, simulate_driving=False, minimum_soc_end=0.8,
                 eta=1.0):
        super().__init__(environment, e_el_max, p_el_max_charge, p_el_max_discharge, soc_init, eta,
                         storage_end_equality=False)
        self._kind = "electricvehicle"
        self._long_id = "EV_" + self._id_string
        self.charging_time_initial = charging_time
        self.ct_pattern = ct_pattern
        self.simulate_driving = simulate_driving
        self.minimum_soc_end = minimum_soc_end

        if charging_time is None:
            # load at night, drive during day
            ts_per_day = int(24 / self.time_slot)
            a = int(ts_per_day / 4)
            b = int(ts_per_day / 2)
            c = ts_per_day - a - b
            charging_time = [1] * a + [0] * b + [1] * c
            self.ct_pattern = 'daily'
        self.charging_time = util.compute_profile(self.timer, charging_time, self.ct_pattern)

        # Contains all starting timesteps of charging periods.
        self.starting_timesteps = np.where(np.diff(charging_time, prepend=0, append=0) == 1)[0]
        # Contains all stopping timesteps of charging periods.
        self.stopping_timesteps = np.where(np.diff(charging_time, prepend=0, append=0) == -1)[0]
        # Contains all length of charging periods in timesteps.
        self.charging_durations = self.stopping_timesteps - self.starting_timesteps
        # Links charging periods to indices in previous arrays.
        self.charging_indices = np.zeros_like(charging_time, dtype=np.int_)
        for i, start in enumerate(self.starting_timesteps):
            self.charging_indices[start:start + self.charging_durations[i]] = np.full(self.charging_durations[i], i)

        # Contains the amount of energy to charge in charging periods.
        self.required_charges = np.full_like(self.charging_durations, self.minimum_soc_end * e_el_max, dtype=np.float64)

        if len(self.charging_time) > 0:
            # In first charging cycle the required charge is determined by the difference between soc_init and
            # minimum_soc_end.
            self.required_charges[0] = (1 - (minimum_soc_end - soc_init)) * e_el_max

            if self.charging_time[-1]:
                # Limit charge at end of simulation horizon to not become infeasible.
                self.required_charges[-1] = min(self.required_charges[-1], self.charging_durations[-1] *
                                                self.time_slot * p_el_max_charge * self.eta_charge)

        if any(self.required_charges > self.charging_durations * self.time_slot * p_el_max_charge *
               self.eta_charge):
            warn("Charging pattern results in infeasible constraints.")
        if self.simulate_driving:
            self.new_var("p_el_drive")

    def populate_model(self, model, mode="convex"):
        """
        Add device block to pyomo ConcreteModel

        Call parent's `populate_model` method. Replace coupling
        constraints from Battery class with coupling constraints
        of EV. Simulate power consumption while driving.

        Parameters
        ----------
        model : pyomo.ConcreteModel
        mode : str, optional
            Specifies which set of constraints to use.

            - `convex`  : Use linear constraints
            - `integer`  : Use integer variables representing discrete control decisions
        """
        super().populate_model(model, mode)
        m = self.model

        if mode == "convex" or "integer":
            # Replace coupling constraints from Battery class
            m.del_component("e_constr")

            # Simulate power consumption while driving
            m.p_el_drive_vars = pyomo.Var(m.t, domain=pyomo.Reals, bounds=(0.0, constants.INFINITY),
                                          initialize=0)

            def p_el_drive_vars_lb_rule(model, t):
                return model.p_el_drive_vars[t] >= 0.0
            m.p_el_drive_vars_lb_constr = pyomo.Constraint(m.t, rule=p_el_drive_vars_lb_rule)

            def p_el_drive_vars_ub_rule(model, t):
                return model.p_el_drive_vars[t] <= constants.INFINITY
            m.p_el_drive_vars_ub_constr = pyomo.Constraint(m.t, rule=p_el_drive_vars_ub_rule)

            # Reset e_el bounds
            def e_el_vars_ev_lb_rule(model, t):
                if t == self.op_horizon-1:
                    return model.e_el_vars[t] >= self.minimum_soc_end * self.e_el_max
                else:
                    return model.e_el_vars[t] >= self.soc_init * self.e_el_max
            m.e_el_vars_ev_lb_constr = pyomo.Constraint(m.t, rule=e_el_vars_ev_lb_rule)

            def e_el_vars_ev_ub_rule(model, t):
                return model.e_el_vars[t] <= self.e_el_max
            m.e_el_vars_ev_ub_constr = pyomo.Constraint(m.t, rule=e_el_vars_ev_ub_rule)

            def e_rule(model, t):
                if self.simulate_driving:
                    delta = (
                        (self.eta_charge * model.p_el_demand_vars[t]
                         - (1.0 / self.eta_discharge) * model.p_el_supply_vars[t]
                         - model.p_el_drive_vars[t])
                        * self.time_slot
                    )
                else:
                    delta = (
                        (self.eta_charge * model.p_el_demand_vars[t]
                         - (1.0 / self.eta_discharge) * model.p_el_supply_vars[t])
                        * self.time_slot
                    )
                e_el_last = model.e_el_vars[t-1] if t >= 1 else model.e_el_init
                return model.e_el_vars[t] == e_el_last + delta
            m.e_constr = pyomo.Constraint(m.t, rule=e_rule)
        else:
            raise ValueError(
                "Mode %s is not implemented by class ElectricVehicle." % str(mode)
            )
        return

    def update_model(self, mode=""):
        m = self.model

        super().update_model(mode)

        timestep = self.timestep
        charging_time = self.charging_time[self.op_slice]
        # Is true if t is before the initial charging period.
        is_initial = not any(self.charging_time[:timestep])
        is_initial_arr = np.empty_like(charging_time, dtype=bool)
        for t in self.op_time_vec:
            if charging_time[t]:
                is_initial = False
            is_initial_arr[t] = is_initial

        p_el_demand_vars_ub_constr_active = m.p_el_demand_vars_ub_constr.active

        def p_el_demand_vars_ub_rule(model, t):
            if charging_time[t]:
                return model.p_el_demand_vars[t] <= self.p_el_max_charge
            else:
                return model.p_el_demand_vars[t] <= 0.0
        try:
            m.del_component(m.p_el_demand_vars_ub_constr)
        except AttributeError:
            pass
        m.p_el_demand_vars_ub_constr = pyomo.Constraint(m.t, rule=p_el_demand_vars_ub_rule)

        if p_el_demand_vars_ub_constr_active is False:
            m.p_el_demand_vars_ub_constr.deactivate()

        p_el_supply_vars_ub_constr_active = m.p_el_supply_vars_ub_constr.active

        def p_el_supply_vars_ub_rule(model, t):
            if charging_time[t]:
                if self.simulate_driving:
                    return model.p_el_supply_vars[t] <= self.p_el_max_discharge
                else:
                    return model.p_el_supply_vars[t] <= 0.0
            else:
                return model.p_el_supply_vars[t] <= 0.0
        try:
            m.del_component(m.p_el_supply_vars_ub_constr)
        except AttributeError:
            pass
        m.p_el_supply_vars_ub_constr = pyomo.Constraint(m.t, rule=p_el_supply_vars_ub_rule)

        if p_el_supply_vars_ub_constr_active is False:
            m.p_el_supply_vars_ub_constr.deactivate()

        p_el_drive_vars_ub_constr_active = m.p_el_drive_vars_ub_constr.active

        def p_el_drive_vars_ub_rule(model, t):
            if charging_time[t]:
                if self.simulate_driving:
                    return model.p_el_drive_vars[t] <= 0.0
                else:
                    return model.p_el_drive_vars[t] <= constants.INFINITY
            else:
                return model.p_el_drive_vars[t] <= constants.INFINITY
        try:
            m.del_component(m.p_el_drive_vars_ub_constr)
        except AttributeError:
            pass
        m.p_el_drive_vars_ub_constr = pyomo.Constraint(m.t, rule=p_el_drive_vars_ub_rule)

        if p_el_drive_vars_ub_constr_active is False:
            m.p_el_drive_vars_ub_constr.deactivate()

        e_el_vars_lb_constr_active = m.e_el_vars_lb_constr.active

        def e_el_vars_lb_rule(model, t):
            if t == 0:
                e_el_vars_lb = self.soc_init * self.e_el_max
            else:
                e_el_vars_lb = model.e_el_vars[t-1]
            if not charging_time[t] and is_initial_arr[t]:
                e_el_vars_lb = self.soc_init * self.e_el_max
            if t + 1 < self.op_horizon:
                if charging_time[t] and not charging_time[t+1]:
                    e_el_vars_lb = self.minimum_soc_end * self.e_el_max
            return model.e_el_vars[t] >= e_el_vars_lb
        try:
            m.del_component(m.e_el_vars_lb_constr)
        except AttributeError:
            pass
        m.e_el_vars_lb_constr = pyomo.Constraint(m.t, rule=e_el_vars_lb_rule)

        if e_el_vars_lb_constr_active is False:
            m.e_el_vars_lb_constr.deactivate()

        e_el_vars_ub_constr_active = m.e_el_vars_ub_constr.active

        def e_el_vars_ub_rule(model, t):
            e_el_vars_ub = self.e_el_max
            if not charging_time[t] and is_initial_arr[t]:
                e_el_vars_ub = self.soc_init * self.e_el_max
            if t + 1 < self.op_horizon:
                if charging_time[t] and not charging_time[t+1]:
                    e_el_vars_ub = self.e_el_max
            return model.e_el_vars[t] <= e_el_vars_ub
        try:
            m.del_component(m.e_el_vars_ub_constr)
        except AttributeError:
            pass
        m.e_el_vars_ub_constr = pyomo.Constraint(m.t, rule=e_el_vars_ub_rule)

        if e_el_vars_ub_constr_active is False:
            m.e_el_vars_ub_constr.deactivate()
        return

    def get_objective(self, coeff=1):
        """
        Objective function for entity level scheduling.

        Return the objective function of the electric vehicle weighted with
        coeff. Quadratic term with additional weights to reward charging the
        vehicle earlier.

        Parameters
        ----------
        coeff : float, optional
            Coefficient for the objective function.

        Returns
        -------
        ExpressionBase :
            Objective function.
        """
        m = self.model
        c = np.array(list(map(lambda x: x+1, range(self.op_horizon))))
        c = c * (coeff * self.op_horizon / sum(c))
        obj = 0
        for t in range(self.op_horizon):
            obj += c[t] * m.p_el_vars[t] * m.p_el_vars[t]
        return obj

    def get_decision_var(self):
        return self.model.p_el_vars



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import pyomo.environ as pyomo
import pycity_base.classes.demand.space_heating as sh

from pycity_scheduling.classes.thermal_entity_heating import ThermalEntityHeating


class SpaceHeating(ThermalEntityHeating, sh.SpaceHeating):
    """
    Extension of pyCity_base class SpaceHeating for scheduling purposes.

    As for all uncontrollable loads, the `p_th_heat_schedule` contains the forecast
    of the load.

    Parameters
    ----------
    environment : Environment
        common to all other objects, includes time and weather instances
    method : int, optional
        - 0 : Provide load curve directly
        - 1 : Use thermal standard load profile
        - 2 : Use ISO 13790 standard to compute thermal load
    loadcurve : numpy.ndarray of float, optional
        load curve for all investigated time steps in [kW]
        requires `method=0`.
    living_area : float, optional
        living area of the apartment in m2
        requires `method=1`
    specific_demand : float, optional
        specific thermal demand of the building in [kWh /(m2*a)]
        requires `method=1`
    profile_type : str, optional
        thermal SLP profile name
        requires `method=1`
        - "HEF" : Single family household
        - "HMF" : Multi family household
        - "GBA" : Bakeries
        - "GBD" : Other services
        - "GBH" : Accomodations
        - "GGA" : Restaurants
        - "GGB" : Gardening
        - "GHA" : Retailers
        - "GHD" : Summed load profile business, trade and services
        - "GKO" : Banks, insurances, public institutions
        - "GMF" : Household similar businesses
        - "GMK" : Automotive
        - "GPD" : Paper and printing
        - "GWA" : Laundries
    zone_parameters : ZoneParameters object, optional
        parameters of the building (floor area, building class, etc.) for `method=2`.
    t_m_init : float, optional
        Initial temperature of the internal heat capacity in [°C] for `method=2`.
    ventilation : array_like, optional
        Ventilation rate in [1/h] for `method=2`.
    t_cooling_set : array_like, optional
        Cooling starts if the room temperature exceeds this value for `method=2`.
    t_heating_set : array_like, optional
        Heating starts if the room temperature drops below this value for `method=2`.
    occupancy : array_like, optional
        Full year occupancy profile for `method=2`.
    appliances : array_like, optional
        Internal gains from electrical appliances in [W] for `method=2`.
    lighting : array_like, optional
        Internal gains from lighting in Watt for `method=2`.

    Notes
    -----
    - The thermal standard load profile is based on the dissertation of Mark Hellwig
      "Entwicklung und Anwendung parametrisierter Standard-Lastprofile",
      TU München, Germany, 2003: http://mediatum.ub.tum.de/doc/601557/601557.pdf (accessed on 2020/09/28)

    - The following constraint is added for removing the bounds from TEH:

    .. math::
        p_{th\\_heat} = load\\_curve
    """

    def __init__(self, environment, method=0, loadcurve=1, living_area=0, specific_demand=0, profile_type='HEF',
                 zone_parameters=None, t_m_init=None, ventilation=0, t_cooling_set=200, t_heating_set=-50, occupancy=0,
                 appliances=0, lighting=0):

        super().__init__(environment, method, loadcurve*1000, living_area, specific_demand, profile_type,
                         zone_parameters, t_m_init, ventilation, t_cooling_set, t_heating_set, occupancy, appliances,
                         lighting)
        self._long_id = "SH_" + self._id_string

        ts = self.timer.time_in_year(from_init=True)
        p = self.loadcurve[ts:ts+self.simu_horizon] / 1000
        self.p_th_heat_schedule = p

    def populate_model(self, model, mode="convex"):
        super().populate_model(model, mode)
        m = self.model
        timestep = self.timestep

        def p_th_heat_vars_rule(model, t):
            return model.p_th_heat_vars[t] == self.p_th_heat_schedule[timestep + t]
        m.p_th_heat_vars_constr = pyomo.Constraint(m.t, rule=p_th_heat_vars_rule)
        return

    def update_model(self, mode=""):
        m = self.model
        timestep = self.timestep

        p_th_heat_vars_constr_active = m.p_th_heat_vars_constr.active

        def p_th_heat_vars_rule(model, t):
            return model.p_th_heat_vars[t] == self.p_th_heat_schedule[timestep + t]
        try:
            m.del_component(m.p_th_heat_vars_constr)
        except AttributeError:
            pass
        m.p_th_heat_vars_constr = pyomo.Constraint(m.t, rule=p_th_heat_vars_rule)

        if p_th_heat_vars_constr_active is False:
            m.p_th_heat_vars_constr.deactivate()
        return

    def new_schedule(self, schedule):
        super().new_schedule(schedule)
        self.copy_schedule(schedule, "default", "p_th_heat")
        return

    def update_schedule(self, mode=""):
        pass

    def reset(self, schedule=None):
        pass



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import warnings

import numpy as np
import pyomo.environ as pyomo

from pycity_scheduling.util.write_schedules import schedule_to_csv, schedule_to_json
from pycity_scheduling.exceptions import SchedulingError


__all__ = [
    'compute_profile',
    'calculate_flexibility_potential',
    'extract_pyomo_value',
    'extract_pyomo_values',
]


def compute_profile(timer, profile, pattern=None):
    """
    Compute a load series profile for an electrical vehicle.

    Parameters
    ----------
    timer : pycity_scheduling.classes.Timer
    profile : array of binaries
        Indicator when electrical vehicle can be charged.

        - `profile[t] == 0`: EV cannot be charged in t
        - `profile[t] == 1`: EV can be charged in t
        It must contain at least one `0` otherwise the model will become
        infeasible. Its length has to be consistent with `pattern`.
    pattern : str, optional
        Define how the `profile` is to be used.

        - `None` : Profile matches simulation horizon.
        - 'daily' : Profile matches one day.
        - 'weekly' : Profile matches one week.

    Returns
    -------
        numpy.ndarray
    """
    if pattern is None:
        if len(profile) != timer.simu_horizon:
            raise ValueError(
                "Length of `profile` does not match `self.simu_horizon`. "
                "Expected: {}, Got: {}"
                .format(timer.simu_horizon, len(profile))
            )
        else:
            return profile
    elif pattern == 'daily':
        ts_per_day = int(86400 / timer.time_discretization)
        if len(profile) != ts_per_day:
            raise ValueError(
                "Length of `profile` does not match one day. Expected: {}, "
                "Got: {}".format(ts_per_day, len(profile))
            )
        else:
            days = int(timer.simu_horizon / ts_per_day) + 2
            ts = timer.time_in_day(from_init=True)
            return np.tile(profile, days)[ts:ts + timer.simu_horizon]
    elif pattern == 'weekly':
        ts_per_week = int(604800 / timer.time_discretization)
        if len(profile) != ts_per_week:
            raise ValueError(
                "Length of `profile` does not match one week. Expected: {}, "
                "Got: {}".format(ts_per_week, len(profile))
            )
        else:
            weeks = int(timer.simu_horizon / ts_per_week) + 2
            ts = timer.time_in_week(from_init=True)
            return np.tile(profile, weeks)[ts:ts + timer.simu_horizon]
    else:
        raise ValueError(
            "Unknown `pattern`: {}. Must be `None`, 'daily' or 'weekly'."
            .format(pattern)
        )


def calculate_flexibility_potential(city_district, algorithm="central", reference_algorithm="stand-alone"):
    """
    Calculate and quantify the operational flexibility potential for a certain city district.


    Parameters
    ----------
    city_district : pycity_scheduling.classes.CityDistrict
        District for which the flexibility potential should be quantified.
    algorithm : str
        Define which algorithm should be used for the flexibility potential quantification purposes.
        Must be one of 'exchange-admm', 'dual-decomposition', 'stand-alone', 'local' or 'central'.
        Default: 'central'.
    reference_algorithm : str
        Define which algorithm should be used as the reference for the flexibility potential quantification purposes.
        Must be one of 'exchange-admm', 'dual-decomposition', 'stand-alone', 'local' or 'central'.
        Default: 'stand-alone'.

    Returns
    -------
    float :
        City district operational flexibility potential in kWh.
    """
    from pycity_scheduling.util.metric import absolute_flexibility_gain
    from pycity_scheduling.algorithms import algorithms

    city_district.copy_schedule(dst="tmp")

    f = algorithms[reference_algorithm](city_district)
    f.solve()
    city_district.copy_schedule("flexibility-potential-quantification-ref")
    f = algorithms[algorithm](city_district)
    f.solve()
    flex = absolute_flexibility_gain(city_district, "flexibility-potential-quantification-ref")

    city_district.copy_schedule(src="tmp")
    return flex


def _known_domains(variable):
    if not variable.is_integer():
        return float
    elif variable.is_binary():
        return bool
    else:
        return int


_numpy_type = {
    float: np.float64,
    int: np.int_,
    bool: np.bool_
}


def extract_pyomo_value(variable, var_type=None, verbose=False):
    """
    Extract a single values out of the pyomo Variable after optimization.

    Parameters
    ----------
    variable : pyomo.Var
        Variable to extract value from.
    var_type : type, optional
        Type with which variable should be stored. Defaults to Domain of pyomo Variable Container or float.

        - float : Store values as floating point numbers.
        - int : Store values as integers (rounding down if necessary).
        - bool : Store values as binary values.
    verbose : bool, optional
        Verbose mode and output.

    Returns
    -------
    float or int or bool:
        Extracted value from the pyomo Variable or the closest value to zero if stale.

    Raises
    ------
    SchedulingError
        If value to extract is not feasible.
    """
    if var_type is None:
        var_type = _known_domains(variable)

    if not hasattr(variable, "stale"):
        raise ValueError("Variable Container does not appear to have been scheduled.")

    if variable.is_indexed():
        raise ValueError("For indexed variables 'extract_pyomo_values' should be used.")

    if variable.stale is True:
        if verbose:
            warnings.warn("Found stale variable {} from optimizer. Make sure that the variable is in active constraints"
                          " and that it is extracted before a new model is optimized. Otherwise, it can not be"
                          " guaranteed, that the value adheres to all constraints.".format(str(variable)))
        # if stale select closest feasible value to zero/variable value
        if variable.value is not None:
            value = variable.value
        else:
            value = 0
        if variable.ub is not None and variable.ub < value:
            value = variable.ub
        elif variable.lb is not None and variable.lb > value:
            value = variable.lb

        if var_type is int:
            if value > 0:
                value = np.ceil(value)
            elif value < 0:
                value = np.floor(value)
        elif var_type is bool:
            if value != 0:
                value = 1

        if (variable.lb is not None and variable.lb > value) or \
           (variable.ub is not None and variable.ub < value):
            raise SchedulingError("Domain and/or bounds of variable render it infeasible.")
        return var_type(value)
    else:
        value = variable.value
    value = var_type(value)
    return value


def extract_pyomo_values(variable, var_type=None, verbose=False):
    """
    Extract values out of the pyomo Variable container after optimization.

    Parameters
    ----------
    variable : pyomo.Var
        Variable container to extract values from.
    var_type : type, optional
        Type with which variable should be stored. Defaults to Domain of pyomo Variable Container or float.

        - float : Store values as floating point numbers.
        - int : Store values as integers (rounding down if necessary).
        - bool : Store values as binary values.
    verbose : bool, optional
        Verbose mode and output.

    Returns
    -------
    numpy.ndarray or float or int or bool:
        If the Variable container is indexed an array containing the extracted values is returned.
        If the Variable container is not indexed returns the single extracted value.

    Raises
    ------
    SchedulingError
        If values to extract are not feasible.
    """

    if not variable.is_constructed():
        raise ValueError("Variable is not constructed.")

    if variable.is_indexed():
        if var_type is None:
            ts = list(_known_domains(v) for v in variable.values())
            if not all(t is ts[0] for t in ts):
                t = float
            else:
                t = ts[0]
        else:
            t = var_type
        dtype = _numpy_type[t]
        values = np.zeros(len(variable), dtype=dtype)
        for i, v in enumerate(variable.values()):
            values[i] = extract_pyomo_value(v, var_type=t, verbose=verbose)
        return values
    else:
        return extract_pyomo_value(variable, var_type, verbose=verbose)



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import matplotlib.pyplot as plt

from pycity_scheduling.classes import *
from pycity_scheduling.algorithms import *
import pycity_scheduling.util.debug as debug


# This is a fundamental tutorial on how to use the pycity_scheduling package.


def main(do_plot=False):
    print("\n\n------ Example 00: Fundamentals ------\n\n")

    # 1) Environment objects:

    # (Almost) every object within pycity_scheduling requires an environment. The environment object holds general data,
    # which is valid for all objects within pycity_scheduling, such as time data, weather data or energy market prices.
    # Therefore, all objects point to an environment. The first step is usually to generate such an environment.

    # Generate a timer object for the environment:
    timer = Timer(step_size=3600, op_horizon=24, initial_date=(2015, 1, 1), initial_time=(0, 0, 0))

    # Generate a weather object for the environment:
    weather = Weather(timer=timer)

    # Generate a price object for the environment:
    price = Prices(timer=timer)

    # Generate the environment object:
    environment = Environment(timer=timer, weather=weather, prices=price)

    # Now there is a distinct environment object with timer, weather and price data.
    # We can use it to access different data of interest.

    # For example, print the current weekday:
    print('Weekday:')
    print(environment.timer.weekday)

    # For example, print the weather forecast for the outdoor temperature (only extract the first 10 timestep values):
    print('\nOutdoor temperature forecast:')
    print(environment.weather.getWeatherForecast(getTAmbient=True)[0][:10])

    # For example, print the energy spot market day-ahead prices:
    print('\nDay-ahead spot market prices on 2015/01/01:')
    print(environment.prices.da_prices)

    # 2) Buildings objects:

    # After defining the environment, different building objects should be created. In pycity_scheduling, buildings
    # represent the different customers of the local energy system / city district under investigation.
    # In general, buildings should also be understood in a more abstract way. For instance, a building object must not
    # necessarily represent a building structure, as it would be the case for a wind energy converter.

    # Create a building object:
    building = Building(environment=environment)

    # This building is assumed to be equipped with a building energy system and one apartment (=single-family house).
    # In general, every building object can, however, hold up to N different apartments (=multi-family house).
    apartment = Apartment(environment=environment)
    bes = BuildingEnergySystem(environment=environment)

    building.addMultipleEntities([apartment, bes])

    # Every apartment usually possesses both electrical and thermal loads:
    # The electrical load is added to the apartment as follows:
    load = FixedLoad(environment=environment, method=1, annual_demand=3000)

    # Print and show the electrical power curve in Watt:
    print('\nElectrical load in Watt:')
    print(load.get_power(currentValues=False))
    plt.plot(load.get_power(currentValues=False))
    plt.xlabel('Time in hours')
    plt.ylabel('Electrical power in Watt (fixed load)')
    plt.title('Fixed load power curve')
    if do_plot:
        plt.show()

    # The thermal load is added to the apartment as follows:
    space_heating = SpaceHeating(environment=environment, method=1, living_area=150, specific_demand=100)

    # Print and show space heating power curve in Watt:
    print('\nSpace heating power curve in Watt:')
    print(space_heating.get_power(currentValues=False))

    plt.plot(space_heating.get_power(currentValues=False))
    plt.xlabel('Time in hours')
    plt.ylabel('Thermal power in Watt (space heating)')
    plt.title('Space heating power curve')
    if do_plot:
        plt.show()

    apartment.addMultipleEntities([load, space_heating])

    # The BuildingEnergySystem (BES) class is a 'container' for all kind of building energy systems (i.e., electrical
    # and/or thermal assets). For example, we can add an electro-thermal heating system (such as a heatpump plus thermal
    # energy storage) and a photovoltaic unit to a building's BES as done below. In pycity_scheduling all BES devices
    # automatically come with basic scheduling models, which include the required Pyomo optimization variables and
    # several optimization constraints.
    eh = HeatPump(environment=environment, p_th_nom=16.0)
    ths = ThermalHeatingStorage(environment=environment, e_th_max=20.0, soc_init=0.5, loss_factor=0)
    pv = Photovoltaic(environment=environment, method=0, peak_power=8.0)

    bes.addMultipleDevices([eh, ths, pv])

    # Verify if the assets were added successfully (method getHasDevice):
    print('\nBES has heatpump? : ', bes.getHasDevices(all_devices=False, heatpump=True)[0])
    print('BES has thermal heating storage? : ', bes.getHasDevices(all_devices=False, ths=True)[0])
    print('BES has photovoltaic? : ', bes.getHasDevices(all_devices=False, pv=True)[0])

    # 3) CityDistrict objects:

    # In pycity_scheduling, a group of buildings form a CityDistrict object. The CityDistrict is the object to be
    # "provided" to a power scheduling algorithm later on. In other word, it encapsulates all buildings together with
    # their local assets and it hence includes all the optimization problem information and data.

    # Create a city district object:
    cd = CityDistrict(environment=environment)

    # Add the building from above to the city district at a certain position/coordinate (x, y).
    cd.addEntity(entity=building, position=[0, 0])

    # Define and add three other buildings:
    for i in range(3):
        heat_demand = SpaceHeating(environment=environment, method=1, living_area=150, specific_demand=100)

        el_load_demand = FixedLoad(environment=environment, method=1, annual_demand=3000)

        pv = Photovoltaic(environment=environment, method=0, peak_power=5.0)
        bl = Boiler(environment=environment, p_th_nom=24.0)

        ap = Apartment(environment=environment)
        ap.addEntity(heat_demand)
        ap.addEntity(el_load_demand)

        bes = BuildingEnergySystem(environment=environment)
        bes.addDevice(pv)
        bes.addDevice(bl)

        bd = Building(environment=environment)
        bd.addEntity(entity=ap)
        bd.addEntity(entity=bes)

        cd.addEntity(entity=bd, position=[0, i])

    # Print the city district information:
    print('\nTotal number of buildings in city district:')
    print(cd.get_nb_of_building_entities())
    print("\nDetailed city district information:")
    debug.print_district(cd, 3)

    # 4) Power scheduling:

    # The final step is to schedule the buildings/assets inside the city district subject to a certain optimization
    # objective, which can be, for example, peak-shaving. The scheduling is then performed by the user by "passing"
    # the city district object to a certain power scheduling algorithm.
    # Here, the central optimization algorithm is used.

    # Set the city district / district operator objective and perform the power scheduling using the central
    # optimization algorithm:
    cd.set_objective("peak-shaving")
    opt = CentralOptimization(cd)
    opt.solve()

    # The scheduling results obtained from the algorithm run can be (temporally) stored as follows:
    cd.copy_schedule("my_central_scheduling")

    # Print and show the scheduling result (city district power values for every time slot within the defined
    # optimization horizon):
    print("\nPower schedule of city district:")
    print(list(cd.p_el_schedule))

    plt.plot(cd.p_el_schedule, drawstyle='steps')
    plt.xlabel('Time in hours')
    plt.ylabel('Electrical power in Kilowatt')
    plt.title('Electrical Power Schedule of CityDistrict')
    if do_plot:
        plt.show()
    return


if __name__ == '__main__':
    # Run example:
    main(do_plot=True)



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import pyomo.environ as pyomo
import matplotlib.pyplot as plt

from pycity_scheduling.classes import (CityDistrict, Building, Photovoltaic, WindEnergyConverter)
from pycity_scheduling.algorithms.algorithm import IterationAlgorithm, DistributedAlgorithm, SolverNode
from pycity_scheduling.solvers import DEFAULT_SOLVER, DEFAULT_SOLVER_OPTIONS


class ExchangeMIQPADMM(IterationAlgorithm, DistributedAlgorithm):
    """Implementation of the Exchange MIQP ADMM algorithm.

    Implements the Exchange MIQP ADMM algorithm described in [1], [2], and [3].

    Parameters
    ----------
    city_district : CityDistrict
    solver : str, optional
        Solver to use for solving (sub)problems.
    solver_options : dict, optional
        Options to pass to calls to the solver. Keys are the name of
        the functions being called and are one of `__call__`, `set_instance_`,
        `solve`.
        `__call__` is the function being called when generating an instance
        with the pyomo SolverFactory. In addition to the options provided,
        `node_ids` is passed to this call containing the IDs of the entities
        being optimized.
        `set_instance` is called when a pyomo Model is set as an instance of
        a persistent solver. `solve` is called to perform an optimization. If
        not set, `save_results` and `load_solutions` may be set to false to
        provide a speedup.
    mode : str, optional
        Specifies which set of constraints to use.
        - `convex`  : Use linear constraints
        - `integer`  : May use non-linear constraints
    eps_exch_primal : float, optional
        Primal stopping criterion for the ADMM algorithm.
    eps_exch_dual : float, optional
        Dual stopping criterion for the ADMM algorithm.
    gamma : float, optional
        Exchange MIQP ADMM scaling parameter
    gamma_incr : float, optional
        Varying scaling parameter scheme increase parameter
    rho : float, optional
        Stepsize for the ADMM algorithm.
    varying_penalty_parameter : bool, optional
        Apply a varying penalty parameter scheme, see [3] - chapter 3.4.1
    tau_incr : float, optional
        Varying penalty parameter scheme increase parameter
    tau_decr : float, optional
        Varying penalty parameter scheme decrease parameter
    mu : float, optional
        Varying penalty parameter scheme conditional change parameter
    max_iterations : int, optional
        Maximum number of ADMM iterations.
    robustness : tuple, optional
        Tuple of two floats. First entry defines how many time steps are
        protected from deviations. Second entry defines the magnitude of
        deviations which are considered.

    References
    ----------
    [1] "Alternating Direction Method of Multipliers for Decentralized
    Electric Vehicle Charging Control" by Jose Rivera, Philipp Wolfrum,
    Sandra Hirche, Christoph Goebel, and Hans-Arno Jacobsen
    Online: https://mediatum.ub.tum.de/doc/1187583/1187583.pdf (accessed on 2020/09/28)

    [2] "A simple effective heuristic for embedded mixed-integer quadratic programming" by Reza Takapoui,
    Nicholas Moehle, Stephen Boyd, and Alberto Bemporad
    Online: https://web.stanford.edu/~boyd/papers/pdf/miqp_admm.pdf (accessed on 2023/09/06)

    [3] "Distributed Optimization and Statistical Learning via the Alternating Direction Method of Multipliers" by
    Stephen Boyd, Neal Parikh, Eric Chu, Borja Peleato, and Jonathan Eckstein
    Online: https://web.stanford.edu/~boyd/papers/pdf/admm_distr_stats.pdf (accessed on 2023/10/09)
    """
    def __init__(self, city_district, solver=DEFAULT_SOLVER, solver_options=DEFAULT_SOLVER_OPTIONS, mode="convex",
                 x_update_mode="constrained", eps_exch_primal=0.01, eps_exch_dual=0.1, gamma=1.0, gamma_incr=1.0,
                 rho=2.0, varying_penalty_parameter=False, tau_incr=2.0, tau_decr=2.0, mu=10.0,
                 max_iterations=10000, robustness=None):
        super(ExchangeMIQPADMM, self).__init__(city_district, solver, solver_options, mode)

        self.x_update_mode = x_update_mode
        self.eps_exch_primal = eps_exch_primal
        self.eps_exch_dual = eps_exch_dual
        self.gamma = gamma
        self.gamma_incr = gamma_incr
        self.rho = rho
        self.varying_penalty_parameter = varying_penalty_parameter
        self.tau_incr = tau_incr
        self.tau_decr = tau_decr
        self.mu = mu
        self.max_iterations = max_iterations
        self.op_horizon = self.city_district.op_horizon
        self.number_constraints = 0
        self.number_binaries = 0

        # Only consider entities of type CityDistrict, Building, Photovoltaic, WindEnergyConverter
        self._entities = [entity for entity in self.entities if
                          isinstance(entity, (CityDistrict, Building, Photovoltaic, WindEnergyConverter))]

        # Create solver nodes for each entity
        self.nodes = [
            SolverNode(solver, solver_options, [entity], mode, robustness=robustness)
            for entity in self._entities
        ]

        # Create pyomo parameters for each entity
        for node, entity in zip(self.nodes, self._entities):
            node.model.gamma_ = pyomo.Param(mutable=True, initialize=self.gamma)
            node.model.rho_ = pyomo.Param(mutable=True, initialize=self.rho)
            node.model.beta = pyomo.Param(mutable=True, initialize=1)
            node.model.xs_ = pyomo.Param(entity.model.t, mutable=True, initialize=0)
            node.model.us = pyomo.Param(entity.model.t, mutable=True, initialize=0)
            node.model.last_p_el_schedules = pyomo.Param(entity.model.t, mutable=True, initialize=0)

        # Additions to the Exchange ADMM algorithm in order to obtain Exchange MIQP ADMM
        self.feasible = False
        self.district_binaries, self.district_bin_values, self.district_x_k_values, \
            self.district_u_k_values = self._get_binaries()
        self.district_equalities_t, self.district_equalities_n, \
            self.district_inequalities_t, self.district_inequalities_n = self._get_constraints()
        self._set_parameters()
        self._add_objective(exchange_admm_obj_terms=True, miqp_admm_obj_terms=True)

    # Function to get the binary variables from the model, to store them and to change their domain to reals
    def _get_binaries(self):
        district_binaries_list = []
        district_bin_values_list = []
        district_x_k_values_list = []
        district_u_k_values_list = []
        for i, node, entity in zip(range(len(self._entities)), self.nodes, self._entities):
            node_binaries_list = []
            node_bin_values_list = []
            node_x_k_values_list = []
            node_u_k_values_list = []
            if i != 0:
                for en in entity.get_all_entities():
                    for variable in en.model.component_objects(pyomo.Var):
                        if variable[0].domain is pyomo.Binary:
                            bin_index_list = []
                            x_k_index_list = []
                            u_k_index_list = []
                            # change domain
                            variable.domain = pyomo.Reals
                            variable.setlb(0)
                            variable.setub(1)
                            node_binaries_list.append(variable)
                            for index in variable:
                                variable[index].fixed = False
                            for t in range(self.op_horizon):
                                x_value_array = np.zeros(self.max_iterations+1)
                                x_k_value_array = np.zeros(self.max_iterations+1)
                                u_k_value_array = np.zeros(self.max_iterations+1)
                                bin_index_list.append(x_value_array)
                                x_k_index_list.append(x_k_value_array)
                                u_k_index_list.append(u_k_value_array)
                            self.number_binaries += self.op_horizon
                            node_bin_values_list.append(bin_index_list)
                            node_x_k_values_list.append(x_k_index_list)
                            node_u_k_values_list.append(u_k_index_list)
            district_binaries_list.append(node_binaries_list)
            district_bin_values_list.append(node_bin_values_list)
            district_x_k_values_list.append(node_x_k_values_list)
            district_u_k_values_list.append(node_u_k_values_list)
        return np.array(district_binaries_list, dtype=object), np.array(district_bin_values_list, dtype=object),\
            np.array(district_x_k_values_list, dtype=object), np.array(district_u_k_values_list, dtype=object)

    def _get_constraints(self):
        # There are time indexed and none indexed constraints. They all need be stored in separate lists.
        district_equalities_list_t = []
        district_equalities_list_n = []
        district_inequalities_list_t = []
        district_inequalities_list_n = []
        for i, node, entity in zip(range(len(self._entities)), self.nodes, self._entities):
            equalities_list_t = []
            inequalities_list_t = []
            equalities_list_n = []
            inequalities_list_n = []
            equality = False
            inequality = False
            none_index = False
            if i != 0:
                for en in entity.get_all_entities():
                    for constraint in en.model.component_objects(pyomo.Constraint):
                        if self.x_update_mode == "unconstrained":
                            constraint.deactivate()
                        for index in constraint:
                            self.number_constraints += 1
                            if index is None:
                                none_index = True
                            # check if the constraint is an equality constraint and write it in the form Ax-b=0
                            # for each index
                            if pyomo.value(constraint[index].lower) == pyomo.value(constraint[index].upper):
                                expr = constraint[index].body - constraint[index].lower
                                constraint[index].set_value(expr == 0)
                                equality = True

                            # if the constraint is not an equality constraint it has to be an inequality constraint;
                            # the next three checks are about to write that constraint in the form Cx-d >= 0
                            elif pyomo.value(constraint[index].upper) is None:
                                expr = constraint[index].body - constraint[index].lower
                                constraint[index].set_value(expr >= 0)
                                inequality = True

                            elif pyomo.value(constraint[index].lower) is None:
                                expr = -constraint[index].body + constraint[index].upper
                                constraint[index].set_value(expr >= 0)
                                inequality = True
                            else:
                                expr = -constraint[index].body + constraint[index].upper
                                expr = constraint[index].body - constraint[index].lower
                                print("WARNING: This constraint has a lb and ub and is hence not deactivated!")
                                inequality = True
                        if equality:
                            if none_index:
                                equalities_list_n.append(constraint)
                                none_index = False
                            else:
                                equalities_list_t.append(constraint)
                            equality = False
                        if inequality:
                            if none_index:
                                inequalities_list_n.append(constraint)
                                none_index = False
                            else:
                                inequalities_list_t.append(constraint)
                            inequality = False
            district_equalities_list_t.append(equalities_list_t)
            district_equalities_list_n.append(equalities_list_n)
            district_inequalities_list_t.append(inequalities_list_t)
            district_inequalities_list_n.append(inequalities_list_n)
        return np.array(district_equalities_list_t, dtype=object), np.array(district_equalities_list_n, dtype=object),\
            np.array(district_inequalities_list_t, dtype=object), np.array(district_inequalities_list_n, dtype=object)

    def _set_parameters(self):
        for i, node, entity in zip(range(len(self._entities)), self.nodes, self._entities):
            # Create parameters for each binary variable
            length = len(self.district_binaries[i])
            if length != 0:
                node.model.bin_set = pyomo.RangeSet(0, length - 1)
                node.model.x_k = pyomo.Param(node.model.bin_set, entity.model.t, mutable=True, initialize=0)
                node.model.u_xk = pyomo.Param(node.model.bin_set, entity.model.t, mutable=True, initialize=0)
            if self.x_update_mode == "unconstrained":
                # Create parameters for each constraint
                length = len(self.district_equalities_t[i])
                if length != 0:
                    node.model.eq_t_set = pyomo.RangeSet(0, length - 1)
                    node.model.u_eq_t = pyomo.Param(node.model.eq_t_set, entity.model.t, mutable=True,
                                                    initialize=0)

                length = len(self.district_equalities_n[i])
                if length != 0:
                    node.model.eq_n_set = pyomo.RangeSet(0, length-1)
                    node.model.u_eq_n = pyomo.Param(node.model.eq_n_set, mutable=True, initialize=0)

                length = len(self.district_inequalities_t[i])
                if length != 0:
                    node.model.ineq_t_set = pyomo.RangeSet(0, length - 1)
                    node.model.u_ineq_t = pyomo.Param(node.model.ineq_t_set, entity.model.t, mutable=True,
                                                      initialize=0)
                    node.model.v_k_t = pyomo.Param(node.model.ineq_t_set, entity.model.t, mutable=True,
                                                   initialize=0)

                length = len(self.district_inequalities_n[i])
                if length != 0:
                    node.model.ineq_n_set = pyomo.RangeSet(0, length - 1)
                    node.model.u_ineq_n = pyomo.Param(node.model.ineq_n_set, mutable=True, initialize=0)
                    node.model.v_k_n = pyomo.Param(node.model.ineq_n_set, mutable=True, initialize=0)
        return

    # Returns True, if no constraint or binary variable violations occur
    def _check_violations(self, print_out=True, detailed_print_out=False):
        constr_counter = 0
        var_counter = 0
        fig_number = 0
        for i, node, entity in zip(range(len(self._entities)), self.nodes, self._entities):
            length = len(self.district_binaries[i])
            if length != 0:
                for x, val, x_k, u_k in zip(self.district_binaries[i], self.district_bin_values[i],
                                            self.district_x_k_values[i], self.district_u_k_values[i]):
                    for t in range(self.op_horizon):
                        if abs(x[t].value) < 0.1 or (0.9 < x[t].value < 1.1):
                            pass
                        else:
                            var_counter += 1
                            if detailed_print_out:
                                x[t].pprint()
                                plot_name = x.name + str(t)
                                plt.figure(fig_number)
                                plt.plot(val[t], label='x')
                                plt.plot(x_k[t], label='x_k')
                                plt.plot(u_k[t], label='u_k')
                                plt.legend()
                                plt.title(plot_name)
                                plt.grid()
                                fig_number += 1

            length = len(self.district_equalities_t[i])
            if length != 0:
                for eq_constr in self.district_equalities_t[i]:
                    for t in range(len(eq_constr)):
                        if abs(pyomo.value(eq_constr[t].body)) > 0.05:
                            if detailed_print_out:
                                print(eq_constr[t].body, " ", pyomo.value(eq_constr[t].body))
                            constr_counter += 1

            length = len(self.district_equalities_n[i])
            if length != 0:
                for eq_constr in self.district_equalities_n[i]:
                    if abs(pyomo.value(eq_constr.body)) > 0.05:
                        if detailed_print_out:
                            print(eq_constr.body, " ", pyomo.value(eq_constr.body))
                        constr_counter += 1

            length = len(self.district_inequalities_t[i])
            if length != 0:
                for ineq_constr in self.district_inequalities_t[i]:
                    for t in range(len(ineq_constr)):
                        if pyomo.value(ineq_constr[t].body) < -0.05:
                            if detailed_print_out:
                                print(ineq_constr[t].body, " ", pyomo.value(ineq_constr[t].body))
                            constr_counter += 1

            length = len(self.district_inequalities_n[i])
            if length != 0:
                for ineq_constr in self.district_inequalities_n[i]:
                    if pyomo.value(ineq_constr.body) < -0.05:
                        if detailed_print_out:
                            print(ineq_constr.body, " ", pyomo.value(ineq_constr.body))
                        constr_counter += 1
        if print_out:
            print("Violated constraints:", constr_counter,  "of in total", self.number_constraints)
            print("Violated binaries:", var_counter, "of in total", self.number_binaries)
        if detailed_print_out:
            plt.show()
        if var_counter == 0 and constr_counter == 0:
            return True
        else:
            return False

    # Returns the objective value
    def _get_objective(self):
        obj_value = 0
        for i, node, entity in zip(range(len(self._entities)), self.nodes, self._entities):
            obj_value += pyomo.value(entity.get_objective())
        return obj_value

    def _add_objective(self, exchange_admm_obj_terms=True, miqp_admm_obj_terms=True):
        for i, node, entity in zip(range(len(self._entities)), self.nodes, self._entities):
            obj = node.model.beta * entity.get_objective()
            for t in range(entity.op_horizon):
                obj += node.model.rho_ / 2 * entity.model.p_el_vars[t] * entity.model.p_el_vars[t]

            # In the following, add the additional expressions to solve the sub-problems by Exchange ADMM.
            if exchange_admm_obj_terms:
                # penalty term is expanded and constant is omitted
                if i == 0:
                    # invert sign of p_el_schedule and p_el_vars (omitted for quadratic term)
                    penalty = [(-node.model.last_p_el_schedules[t] - node.model.xs_[t] - node.model.us[t])
                               for t in range(entity.op_horizon)]
                    for t in range(entity.op_horizon):
                        obj += node.model.rho_ * penalty[t] * entity.model.p_el_vars[t]
                else:
                    penalty = [(-node.model.last_p_el_schedules[t] + node.model.xs_[t] + node.model.us[t])
                               for t in range(entity.op_horizon)]
                    for t in range(entity.op_horizon):
                        obj += node.model.rho_ * penalty[t] * entity.model.p_el_vars[t]

            # In the following, add the additional expressions to solve the sub-problems by MIQP ADMM.
            if miqp_admm_obj_terms:
                # binary variables contribution
                length = len(self.district_binaries[i])
                if length != 0:
                    for x, k in zip(self.district_binaries[i], range(length)):
                        obj += node.model.rho_ / 2 * sum((x[t] - node.model.x_k[k, t] + node.model.gamma_ *
                                                          node.model.u_xk[k, t]) ** 2 for t in range(self.op_horizon))

                if self.x_update_mode == "unconstrained":
                    # add the contributions of the constraints (time and none indexed)
                    length = len(self.district_equalities_t[i])
                    if length != 0:
                        for eq_constr, k in zip(self.district_equalities_t[i], range(length)):
                            obj += node.model.rho_ / 2 * sum((eq_constr[t].body + node.model.u_eq_t[k, t]) ** 2
                                                             for t in range(len(eq_constr)))

                    length = len(self.district_equalities_n[i])
                    if length != 0:
                        for eq_constr, k in zip(self.district_equalities_n[i], range(length)):
                            obj += node.model.rho_ / 2 * (eq_constr.body + node.model.u_eq_n[k]) ** 2

                    length = len(self.district_inequalities_t[i])
                    if length != 0:
                        for ineq_constr, k in zip(self.district_inequalities_t[i], range(length)):
                            obj += node.model.rho_ / 2 * sum((ineq_constr[t].body + node.model.u_ineq_t[k, t] -
                                                              node.model.v_k_t[k, t]) ** 2
                                                             for t in range(len(ineq_constr)))

                    length = len(self.district_inequalities_n[i])
                    if length != 0:
                        for ineq_constr, k in zip(self.district_inequalities_n[i], range(length)):
                            obj += node.model.rho_ / 2 * (ineq_constr.body + node.model.u_ineq_n[k] -
                                                          node.model.v_k_n[k]) ** 2

            # if we want to redefine the objective for a certain node, then we should first reset the old objective
            try:
                node.model.del_component(node.model.o)
            except AttributeError:
                pass
            node.model.o = pyomo.Objective(expr=obj)
        return

    def _presolve(self, full_update, beta, robustness, debug):
        results, params = super()._presolve(full_update, beta, robustness, debug)

        for node, entity in zip(self.nodes, self._entities):
            node.model.beta = self._get_beta(params, entity)
            if full_update:
                node.full_update(robustness)
        results["r_norms"] = []
        results["s_norms"] = []
        results["gamma_value"] = []
        results["rho_value"] = []
        return results, params

    def _is_last_iteration(self, results, params, debug):
        if super(ExchangeMIQPADMM, self)._is_last_iteration(results, params, debug):
            self._check_violations()
            return True
        if results["r_norms"][-1] <= self.eps_exch_primal and results["s_norms"][-1] <= self.eps_exch_dual:
            if self.feasible:
                self._check_violations()
                return True
            else:
                return False

    def _iteration(self, results, params, debug):
        super(ExchangeMIQPADMM, self)._iteration(results, params, debug)
        op_horizon = self._entities[0].op_horizon

        # fill parameters if not already present
        if "p_el" not in params:
            params["p_el"] = np.zeros((len(self._entities), op_horizon))
        if "x_" not in params:
            params["x_"] = np.zeros(op_horizon)
        if "u" not in params:
            params["u"] = np.zeros(op_horizon)
        last_u = params["u"]
        last_p_el = params["p_el"]
        last_x_ = params["x_"]

        # --------------------------
        # 1) optimize all entities
        # --------------------------

        to_solve_nodes = []
        variables = []
        for i, node, entity in zip(range(len(self._entities)), self.nodes, self._entities):
            for t in range(op_horizon):
                node.model.last_p_el_schedules[t] = last_p_el[i][t]
                node.model.xs_[t] = last_x_[t]
                node.model.us[t] = last_u[t]
            node.obj_update()
            to_solve_nodes.append(node)
            variables.append([entity.model.p_el_vars[t] for t in range(op_horizon)])
        self._solve_nodes(results, params, to_solve_nodes, variables=variables, debug=debug)

        # ----------------------------------------------
        # 2) incentive signal update (Exchange ADMM)
        # ----------------------------------------------

        p_el_schedules = np.stack([entity.p_el_schedule for entity in self._entities], axis=0)
        x_ = (-p_el_schedules[0] + sum(p_el_schedules[1:])) / len(self._entities)

        results["r_norms"].append(np.sqrt(len(self._entities)) * np.linalg.norm(x_))

        s = np.zeros_like(p_el_schedules)
        s[0] = - self.rho * (-p_el_schedules[0] + last_p_el[0] + last_x_ - x_)
        for i in range(1, len(self._entities)):
            s[i] = - self.rho * (p_el_schedules[i] - last_p_el[i] + last_x_ - x_)
        results["s_norms"].append(np.linalg.norm(s.flatten()))

        results["obj_value"].append(self._get_objective())

        # To check the latest solution for feasibility, fix all binary vars at either 0 or 1 and try to recalculate the
        # schedule based on an additional optimization. Of course, there are other and possibly more
        # straightforward ways to check for feasibility. This will be future work.
        if ((results["r_norms"][-1] <= self.eps_exch_primal) and (results["s_norms"][-1] <= self.eps_exch_dual)) or \
                (results["iterations"][-1] >= self.max_iterations):
            if results["iterations"][-1] >= self.max_iterations:
                print("Iteration " + str(results["iterations"][-1]) + ": " +
                      "Maximum iteration limit reached. Checking the current solution for feasibility.")
            else:
                print("Iteration " + str(results["iterations"][-1]) + ": " +
                      "Stopping criteria satisfied. Checking the current solution for feasibility.")
            self._add_objective(exchange_admm_obj_terms=True, miqp_admm_obj_terms=False)
            self.fix_variables()
            if self.x_update_mode == "unconstrained":
                self.activate_constraints()
            self.feasible = True
            for i, node, entity in zip(range(len(self._entities)), self.nodes, self._entities):
                to_solve_nodes = []
                variables = []
                node.obj_update()
                node.constr_update()
                to_solve_nodes.append(node)
                variables.append([entity.model.p_el_vars[t] for t in range(op_horizon)])
                try:
                    self._solve_nodes({}, {}, to_solve_nodes, variables=variables, debug=False)
                except:
                    self.feasible = False
            if self.feasible:
                print("Success. The solution is feasible!")
                self.release_variables()
                if self.x_update_mode == "unconstrained":
                    self.deactivate_constraints()
            else:
                print("Failure. The solution is infeasible, because at least one subsystem is (still) infeasible!")
                if results["iterations"][-1] >= self.max_iterations:
                    print("Attention: The final solution could be infeasible!")
                else:
                    self._add_objective(exchange_admm_obj_terms=True, miqp_admm_obj_terms=True)
                    self.release_variables()
                    if self.x_update_mode == "unconstrained":
                        self.deactivate_constraints()
                    to_solve_nodes = []
                    variables = []
                    for i, node, entity in zip(range(len(self._entities)), self.nodes, self._entities):
                        node.obj_update()
                        node.constr_update()
                        to_solve_nodes.append(node)
                        variables.append([entity.model.p_el_vars[t] for t in range(op_horizon)])
                    self._solve_nodes({}, {}, to_solve_nodes, variables=variables, debug=False)
                    # vary the gamma parameter:
                    self.gamma = min(float(self.gamma_incr)*self.gamma, 10e9)
                    for node, entity in zip(self.nodes, self._entities):
                        node.model.gamma_ = self.gamma
        results["gamma_value"].append(self.rho)

        # ----------------------------------------
        # 3) incentive signal update (MIQP ADMM)
        # ----------------------------------------

        # update all MIQP ADMM parameters (first the constraints then the variables)
        for i, node, entity in zip(range(len(self._entities)), self.nodes, self._entities):
            if self.x_update_mode == "unconstrained":
                length = len(self.district_inequalities_t[i])
                if length != 0:
                    for ineq_constr, k in zip(self.district_inequalities_t[i], range(length)):
                        for t in range(len(ineq_constr)):
                            v_update = pyomo.value(ineq_constr[t].body) + node.model.u_ineq_t[k, t].value
                            if v_update > 0:
                                node.model.v_k_t[k, t] = v_update
                            else:
                                node.model.v_k_t[k, t] = 0

                length = len(self.district_inequalities_n[i])
                if length != 0:
                    for ineq_constr, k in zip(self.district_inequalities_n[i], range(length)):
                        v_update = pyomo.value(ineq_constr.body) + node.model.u_ineq_n[k].value
                        if v_update > 0:
                            node.model.v_k_n[k] = v_update
                        else:
                            node.model.v_k_n[k] = 0

                length = len(self.district_inequalities_t[i])
                if length != 0:
                    for ineq_constr, k in zip(self.district_inequalities_t[i], range(length)):
                        for t in range(len(ineq_constr)):
                            node.model.u_ineq_t[k, t] = node.model.u_ineq_t[k, t].value + \
                                                        pyomo.value(ineq_constr[t].body) - \
                                                        node.model.v_k_t[k, t].value

                length = len(self.district_inequalities_n[i])
                if length != 0:
                    for ineq_constr, k in zip(self.district_inequalities_n[i], range(length)):
                        node.model.u_ineq_n[k] = node.model.u_ineq_n[k].value + pyomo.value(ineq_constr.body) - \
                                                    node.model.v_k_n[k].value

                length = len(self.district_equalities_t[i])
                if length != 0:
                    for eq_constr, k in zip(self.district_equalities_t[i], range(length)):
                        for t in range(len(eq_constr)):
                            node.model.u_eq_t[k, t] = node.model.u_eq_t[k, t].value + pyomo.value(eq_constr[t].body)

                length = len(self.district_equalities_n[i])
                if length != 0:
                    for eq_constr, k in zip(self.district_equalities_n[i], range(length)):
                        node.model.u_eq_n[k] = node.model.u_eq_n[k].value + pyomo.value(eq_constr.body)

            # x_k update
            length = len(self.district_binaries[i])
            if length != 0:
                for x, k in zip(self.district_binaries[i], range(length)):
                    for t in range(op_horizon):
                        # Integer rounding:
                        node.model.x_k[k, t] = round(x[t].value + node.model.u_xk[k, t].value)
                        # Binary rounding:
                        # x_update = abs(x[t].value + node.model.u_xk[k, t].value)
                        # if x_update >= 0.5:
                        #    node.model.x_k[k, t] = 1
                        # else:
                        #    node.model.x_k[k, t] = 0
                        self.district_x_k_values[i][k][t][results["iterations"][-1]-1] = node.model.x_k[k, t].value

            # u_xk update
            length = len(self.district_binaries[i])
            if length != 0:
                for x, x_val, u_k, k in zip(self.district_binaries[i], self.district_bin_values[i],
                                            self.district_u_k_values[i], range(length)):
                    for t in range(op_horizon):
                        node.model.u_xk[k, t] = node.model.u_xk[k, t].value + x[t].value - \
                                                node.model.x_k[k, t].value
                        x_val[t][results["iterations"][-1]-1] = x[t].value
                        u_k[t][results["iterations"][-1]-1] = node.model.u_xk[k, t].value

        # if desired, vary the penalty parameter
        if self.varying_penalty_parameter:
            if results["r_norms"][-1] > self.mu * results["s_norms"][-1]:
                self.rho *= float(self.tau_incr)
                for node, entity in zip(self.nodes, self._entities):
                    node.model.rho_ = self.rho
            elif results["s_norms"][-1] > self.mu * results["r_norms"][-1]:
                self.rho /= float(self.tau_decr)
                for node, entity in zip(self.nodes, self._entities):
                    node.model.rho_ = self.rho
        results["rho_value"].append(self.rho)

        # save parameters for another iteration
        params["p_el"] = p_el_schedules
        params["x_"] = x_
        params["u"] += x_
        return

    def fix_variables(self):
        for i, node, entity in zip(range(len(self._entities)), self.nodes, self._entities):
            length = len(self.district_binaries[i])
            if length != 0:
                for x, k in zip(self.district_binaries[i], range(length)):
                    for t in range(self.op_horizon):
                        x_k = node.model.x_k[k, t].value
                        x[t].fix(x_k)
        return

    def release_variables(self):
        for i, node, entity in zip(range(len(self._entities)), self.nodes, self._entities):
            length = len(self.district_binaries[i])
            if length != 0:
                for x, k in zip(self.district_binaries[i], range(length)):
                    for t in range(self.op_horizon):
                        x[t].unfix()
        return

    def activate_constraints(self):
        for i, node, entity in zip(range(len(self._entities)), self.nodes, self._entities):
            if i != 0:
                for en in entity.get_all_entities():
                    for constraint in en.model.component_objects(pyomo.Constraint):
                        constraint.activate()
        return

    def deactivate_constraints(self):
        for i, node, entity in zip(range(len(self._entities)), self.nodes, self._entities):
            if i != 0:
                for en in entity.get_all_entities():
                    for constraint in en.model.component_objects(pyomo.Constraint):
                        constraint.deactivate()
        return

    def print_model(self):
        for i, node, entity in zip(range(len(self._entities)), self.nodes, self._entities):
            for en in entity.get_all_entities():
                for constraint in en.model.component_objects(pyomo.Constraint):
                    constraint.pprint()
                for x in self.district_binaries[i]:
                    x.pprint()



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import pyomo.environ as pyomo
import pycity_base.classes.supply.boiler as bl

from pycity_scheduling.classes.thermal_entity_heating import ThermalEntityHeating
from pycity_scheduling.util.generic_constraints import LowerActivationLimit


class Boiler(ThermalEntityHeating, bl.Boiler):
    """
    Extension of pyCity_base class Boiler for scheduling purposes.

    Parameters
    ----------
    environment : pycity_scheduling.classes.Environment
        Common to all other objects. Includes time and weather instances.
    p_th_nom : float
        Nominal heat output in [kW].
    eta : float, optional
        Efficiency of the gas boiler (without unit).
    lower_activation_limit : float, optional (only adhered to in integer mode)
        Must be in [0, 1]. Lower activation limit of the gas boiler as a
        percentage of the rated power. When the gas boiler is in operation, its
        power must be zero or between the lower activation limit and its
        rated power.

        - `lower_activation_limit = 0`: Linear behavior
        - `lower_activation_limit = 1`: Two-point controlled
    min_off_time : int, optional (only adhered to in integer mode)
        Minimum number of consecutive time steps the boiler must remain turned off, once it has been decided that it
        should turn off subject to 'lower_activation_limit'. Defaults to '1', i.e., not limited to a consecutive number
        of time steps.
    min_on_time : int, optional (only adhered to in integer mode)
        Minimum number of consecutive time steps the boiler must remain turned on, once it has been decided that it
        should turn on subject to 'lower_activation_limit'. Defaults to '1', i.e., not limited to a consecutive number
        of time steps.

    Notes
    -----
    - Gas boilers offer sets of constraints for operation. In the `convex` mode the
      following constraints and bounds are generated by the boiler:

    .. math::
        0 \\geq p_{th\\_heat} \\geq -p_{th\\_nom}

    - See also:
        - pycity_scheduling.util.generic_constraints.LowerActivationLimit:
          Generates additional constraints for the `lower_activation_limit` in `integer` mode.
    """

    def __init__(self, environment, p_th_nom, eta=1, lower_activation_limit=0.0, min_off_time=1, min_on_time=1):
        # Flow temperature of 55 C
        super().__init__(environment, 1000*p_th_nom, eta, 55, lower_activation_limit)
        self._long_id = "BL_" + self._id_string
        self.p_th_nom = p_th_nom
        self.min_off_time = min_off_time
        self.min_on_time = min_on_time
        self.activation_constr = None

    def populate_model(self, model, mode="convex"):
        """
        Add device block to pyomo ConcreteModel

        Call parent's `populate_model` method and set variables upper bounds
        to `self.p_th_nom`.

        Parameters
        ----------
        model : pyomo.ConcreteModel
        mode : str, optional
            Specifies which set of constraints to use.

            - `convex`  : Use linear constraints
            - `integer`  : Use integer variables representing discrete control decisions
        """
        if self.activation_constr is None:
            self.activation_constr = LowerActivationLimit(self, "p_th_heat", self.lower_activation_limit,
                                                          -self.p_th_nom, self.min_off_time, self.min_on_time)

        super().populate_model(model, mode)
        m = self.model

        if mode == "convex" or mode == "integer":
            def p_th_heat_vars_lb_rule(model, t):
                return model.p_th_heat_vars[t] >= -self.p_th_nom
            m.p_th_heat_vars_lb_constr = pyomo.Constraint(m.t, rule=p_th_heat_vars_lb_rule)

            def p_th_heat_vars_ub_rule(model, t):
                return model.p_th_heat_vars[t] <= 0.0
            m.p_th_heat_vars_ub_constr = pyomo.Constraint(m.t, rule=p_th_heat_vars_ub_rule)

            self.activation_constr.apply(m, mode)
        else:
            raise ValueError(
                "Mode %s is not implemented by class Boiler." % str(mode)
            )
        return

    def get_objective(self, coeff=1):
        """
        Objective function for entity level scheduling.

        Return the objective function of the boiler weighted with `coeff`.
        Sum of `p_th_heat`.

        Parameters
        ----------
        coeff : float, optional
            Coefficient for the objective function.

        Returns
        -------
        ExpressionBase :
            Objective function.
        """
        return coeff * sum(self.model.p_th_heat_vars[t] for t in range(self.op_horizon))

    def get_decision_var(self):
        return self.model.p_th_heat_vars



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import os
import sys
import numpy as np


class MPIInterface:
    """
    Supporting class that handles interactions with the Message Passing Interface (MPI) using module mpi4py.
    MPI is used to perform parallel computations on loosely coupled machines.
    """

    def __init__(self):
        self.mpi = None
        self.mpi_comm = None
        self.mpi_rank = 0
        self.mpi_size = 1
        try:
            from mpi4py import MPI
            self.mpi = MPI
            self.mpi_comm = MPI.COMM_WORLD
            self.mpi_rank = self.mpi_comm.Get_rank()
            self.mpi_size = self.mpi_comm.Get_size()
        except:
            pass

    def get_rank(self):
        """
        Return the rank of the current MPI process.
        """
        return self.mpi_rank

    def get_size(self):
        """
        Return the size, i.e., number, of MPI processes.
        """
        return self.mpi_size

    def get_comm(self):
        """
        Return the MPI COMM interface.
        """
        return self.mpi_comm

    def get_mpi_process_range(self, n):
        """
        Determine which MPI processes are responsible for which elements in an array.

        Parameters
        ----------
        n : int
            Size of the array.

        Returns
        ----------
        mpi_process_range : np.array
            Array, in which the number inside an entry corresponds to the responsible MPI rank.
        """
        if self.get_size() > n:
            mpi_process_range = np.array([i for i in range(n)])
        elif self.get_size() < n:
            if self.get_size() == 1:
                mpi_process_range = np.array([0 for i in range(n)])
            else:
                a, b = divmod(n - 1, self.get_size() - 1)
                mpi_process_range = np.repeat(np.array([i for i in range(1, self.get_size())]), a)
                for i in range(b):
                    mpi_process_range = np.append(mpi_process_range, i + 1)
                mpi_process_range = np.concatenate([[0], mpi_process_range])
        else:
            mpi_process_range = np.array([i for i in range(n)])
        mpi_process_range = np.sort(mpi_process_range)
        return mpi_process_range

    def disable_multiple_printing(self, stdout=True, stderr=True):
        """
        Turn off printing for all MPI processes with MPI rank other than 0 and always flush prints for rank 0.

        Parameters
        ----------
        stdout : bool
            Turn off printing for stdout.
        stderr : bool
            Turn off printing for stderr.
        """
        sys.stdout = UnbufferedPrint(sys.stdout)
        sys.stderr = UnbufferedPrint(sys.stderr)
        if self.mpi is not None and self.mpi_size > 1:
            if self.mpi_rank > 0:
                if stdout:
                    sys.stdout = open(os.devnull, 'w')
                if stderr:
                    sys.stderr = open(os.devnull, 'w')


class UnbufferedPrint(object):
    def __init__(self, stream):
        self.stream = stream

    def write(self, data):
        self.stream.write(data)
        self.stream.flush()

    def writelines(self, data):
        self.stream.writelines(data)
        self.stream.flush()

    def __getattr__(self, attr):
        return getattr(self.stream, attr)



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
from pyomo.opt import TerminationCondition
import pyomo.solvers.plugins.solvers as Solvers


def analyze_model(model, optimizer, result, options={}):
    """
    Analyze a model which is not optimal.

    Parameters
    ----------
    model : pyomo.ConcreteModel
        Model with `status != OPTIMAL`
    optimizer : OptSolver
        The solver that was used for optimization and is used for analyzes
    result: SolverResults
        The not optimal result that was returned by the solver
    options : str, optional
        Options which should be passed to the solver when analyzing
    """
    if result.solver.termination_condition in \
            [TerminationCondition.infeasibleOrUnbounded, TerminationCondition.infeasible] and \
            (isinstance(optimizer, (Solvers.GUROBI.GUROBI,
                                    Solvers.gurobi_persistent.GurobiPersistent,
                                    Solvers.gurobi_direct.GurobiDirect,
                                    Solvers.GUROBI.GUROBISHELL))):

        options["dualreductions"] = 0
        if isinstance(optimizer, Solvers.gurobi_persistent.GurobiPersistent):
            optimizer.set_instance(model, symbolic_solver_labels=True)
            result = optimizer.solve(options=options, tee=True)
            if result.solver.termination_condition == TerminationCondition.infeasible:
                options["ResultFile"] = "model.ilp"
                optimizer.solve(options=options, tee=True)
                print("IIS written to 'model.ilp'.")
        else:
            result = optimizer.solve(model, options=options, tee=True, symbolic_solver_labels=True)
            if result.solver.termination_condition == TerminationCondition.infeasible:
                options["ResultFile"] = "model.ilp"
                optimizer.solve(model, options=options, tee=True, symbolic_solver_labels=True)
                print("IIS written to 'model.ilp'.")
    status = result.solver.status
    condition = result.solver.termination_condition
    print("Model status is {}.".format(status))
    print("Model condition is {}.".format(condition))


def print_district(cd, lvl=1):
    """
    Hierarchically print a city district.

    Parameters
    ----------
    cd : pycity_scheduling.classes.CityDistrict
    lvl : int, optional
        - `0` : Only print city district.
        - `1` : Only print city district and buildings.
        - `2` : Print city district, buildings and all their devices.
    """
    print(cd)
    if lvl < 1:
        return
    for bd in cd.get_lower_entities():
        print("\t{}".format(bd))
        if lvl < 2:
            continue
        if bd.has_bes:
            for e in bd.bes.get_lower_entities():
                print("\t\t{}".format(e))
        if len(bd.apartments) == 1:
            for e in bd.apartments[0].get_lower_entities():
                print("\t\t{}".format(e))
        elif len(bd.apartments) > 1:
            for ap in bd.apartments:
                print("\t\t{}".format(ap))
                for e in ap.get_lower_entities():
                    print("\t\t\t{}".format(e))



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


from .stand_alone_optimization_algorithm import StandAlone
from .local_optimization_algorithm import LocalOptimization
from .exchange_admm_algorithm import ExchangeADMM
from .exchange_admm_algorithm_mpi import ExchangeADMMMPI
from .exchange_admm_algorithm_ray import ExchangeADMMRay
from .central_optimization_algorithm import CentralOptimization
from .dual_decomposition_algorithm import DualDecomposition
from .dual_decomposition_algorithm_mpi import DualDecompositionMPI
from .dual_decomposition_algorithm_ray import DualDecompositionRay
from .exchange_miqp_admm_algorithm import ExchangeMIQPADMM
from .exchange_miqp_admm_algorithm_mpi import ExchangeMIQPADMMMPI
from .exchange_miqp_admm_algorithm_ray import ExchangeMIQPADMMRay
from .derivative_free_aladin_algorithm import DerivativeFreeALADIN
from .derivative_free_aladin_algorithm_ray import DerivativeFreeALADINRay
from .derivative_free_aladin_algorithm_ray_Hvet import DerivativeFreeALADINRay_Hvet


__all__ = [
    'StandAlone',
    'LocalOptimization',
    'ExchangeADMM',
    'ExchangeADMMMPI',
    'ExchangeADMMRay',
    'CentralOptimization',
    'DualDecomposition',
    'DualDecompositionMPI',
    'DualDecompositionRay',
    'ExchangeMIQPADMM',
    'ExchangeMIQPADMMMPI',
    'ExchangeMIQPADMMRay',
    'DerivativeFreeALADIN',
    'DerivativeFreeALADINRay',
    'DerivativeFreeALADINRay_Hvet',
    'algorithm',
    'algorithm_ray',
    'algorithms',
]


algorithms = {
    'stand-alone': StandAlone,
    'local': LocalOptimization,
    'exchange-admm': ExchangeADMM,
    'exchange-admm-mpi': ExchangeADMMMPI,
    'exchange-admm-ray': ExchangeADMMRay,
    'central': CentralOptimization,
    'dual-decomposition': DualDecomposition,
    'dual-decomposition-mpi': DualDecompositionMPI,
    'dual-decomposition-ray': DualDecompositionRay,
    'exchange-miqp-admm': ExchangeMIQPADMM,
    'exchange-miqp-admm-mpi': ExchangeMIQPADMMMPI,
    'exchange-miqp-admm-ray': ExchangeMIQPADMMRay,
    'derivative-free-aladin': DerivativeFreeALADIN,
    'derivative-free-aladin-ray': DerivativeFreeALADINRay,
    'derivative-free-aladin-ray-Hvet': DerivativeFreeALADINRay_Hvet
}



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import matplotlib.pyplot as plt
from matplotlib import gridspec

import pycity_scheduling.util.factory as factory
from pycity_scheduling.classes import *
from pycity_scheduling.algorithms import *


# This is a simple power scheduling example to demonstrate the difference between the pycity_scheduling's convex and
# integer (MIP) optimization models using the central algorithm.


def main(do_plot=False):
    print("\n\n------ Example 23: Scheduling Convex vs. Integer Mode ------\n\n")

    # Scheduling will be performed for a typical winter day within the annual heating period:
    env = factory.generate_standard_environment(step_size=3600, op_horizon=24, mpc_horizon=None, mpc_step_width=None,
                                                initial_date=(2010, 2, 10), initial_time=(0, 0, 0))

    # City district / district operator objective is peak-shaving:
    cd = CityDistrict(environment=env, objective='peak-shaving')

    # Building equipped with space heating, electric heater, thermal energy storage, and photovoltaic unit.
    # Objective is self-consumption:
    bd1 = Building(environment=env, objective='self-consumption')
    cd.addEntity(entity=bd1, position=[0, 0])
    ap1 = Apartment(environment=env)
    bd1.addEntity(ap1)
    sh1 = SpaceHeating(environment=env, method=1, living_area=120.0, specific_demand=85.9, profile_type='HEF')
    ap1.addEntity(sh1)
    bes1 = BuildingEnergySystem(environment=env)
    bd1.addEntity(bes1)
    eh1 = ElectricHeater(environment=env, p_th_nom=12.0, lower_activation_limit=0.5)
    bes1.addDevice(eh1)
    ths1 = ThermalHeatingStorage(environment=env, e_th_max=24.0)
    bes1.addDevice(ths1)
    pv1 = Photovoltaic(environment=env, method=1, peak_power=25.0)
    bes1.addDevice(pv1)

    # First, perform the power scheduling using convex models for the electrical appliances:
    opt = CentralOptimization(cd, mode='convex')
    opt.solve()
    cd.copy_schedule("convex_schedule")

    # Plot the convex schedules:
    plot_time = list(range(env.timer.timesteps_used_horizon))

    gs = gridspec.GridSpec(5, 1)
    ax0 = plt.subplot(gs[0])
    ax0.plot(plot_time, cd.p_el_schedule)
    plt.xlim((0, env.timer.timesteps_used_horizon - 1))
    plt.ylim([-15, 15])
    plt.grid()
    plt.title("Convex Schedules")
    plt.ylabel("District [kW]")

    ax1 = plt.subplot(gs[1], sharex=ax0)
    ax1.plot(plot_time, pv1.p_el_schedule)
    plt.xlim((0, env.timer.timesteps_used_horizon - 1))
    plt.grid()
    plt.ylabel("PV [kW]")

    ax2 = plt.subplot(gs[2], sharex=ax0)
    ax2.plot(plot_time, sh1.p_th_heat_schedule)
    plt.xlim((0, env.timer.timesteps_used_horizon - 1))
    plt.grid()
    plt.ylabel("Space Heating Demand [kW]")

    ax3 = plt.subplot(gs[3], sharex=ax0)
    ax3.plot(plot_time, eh1.p_el_schedule)
    plt.xlim((0, env.timer.timesteps_used_horizon - 1))
    plt.grid()
    plt.ylabel("El. Heater [kW]")

    ax4 = plt.subplot(gs[4], sharex=ax0)
    ax4.plot(plot_time, ths1.e_th_heat_schedule)
    plt.xlim((0, env.timer.timesteps_used_horizon - 1))
    plt.grid()
    plt.ylabel("THS SoC [kWh]")

    plt.xlabel("Time", fontsize=12)

    if do_plot:
        figManager = plt.get_current_fig_manager()
        if hasattr(figManager, "window"):
            figManagerWindow = figManager.window
            if hasattr(figManagerWindow, "state"):
                figManager.window.state("zoomed")
        plt.show()

    # Second, perform the power scheduling using integer models for the electrical appliances:
    opt = CentralOptimization(cd, mode='integer')
    opt.solve()
    cd.copy_schedule("integer_schedule")

    # Plot the integer schedules:
    gs = gridspec.GridSpec(5, 1)
    ax0 = plt.subplot(gs[0])
    ax0.plot(plot_time, cd.p_el_schedule)
    plt.xlim((0, env.timer.timesteps_used_horizon - 1))
    plt.ylim([-15, 15])
    plt.grid()
    plt.title("Integer Schedules")
    plt.ylabel("District [kW]")

    ax1 = plt.subplot(gs[1], sharex=ax0)
    ax1.plot(plot_time, pv1.p_el_schedule)
    plt.xlim((0, env.timer.timesteps_used_horizon - 1))
    plt.grid()
    plt.ylabel("PV [kW]")

    ax2 = plt.subplot(gs[2], sharex=ax0)
    ax2.plot(plot_time, sh1.p_th_heat_schedule)
    plt.xlim((0, env.timer.timesteps_used_horizon - 1))
    plt.grid()
    plt.ylabel("Space Heating Demand [kW]")

    ax3 = plt.subplot(gs[3], sharex=ax0)
    ax3.plot(plot_time, eh1.p_el_schedule)
    plt.xlim((0, env.timer.timesteps_used_horizon - 1))
    plt.grid()
    plt.ylabel("El. Heater [kW]")

    ax4 = plt.subplot(gs[4], sharex=ax0)
    ax4.plot(plot_time, ths1.e_th_heat_schedule)
    plt.xlim((0, env.timer.timesteps_used_horizon - 1))
    plt.grid()
    plt.ylabel("THS SoC [kWh]")

    plt.xlabel("Time", fontsize=12)

    if do_plot:
        figManager = plt.get_current_fig_manager()
        if hasattr(figManager, "window"):
            figManagerWindow = figManager.window
            if hasattr(figManagerWindow, "state"):
                figManager.window.state("zoomed")
        plt.show()
    return


# Conclusions:
# In the convex case, the peak-shaving objective can be fully satisfied, which becomes evident from the "flat" power
# profile of the obtained city district power curve. To perfectly balance the power demand and supply in the district,
# the flexibility from the thermal heating storage unit is exploited together with the capability to operate the
# electric heater at any operation point between 0% and 100% of its nominal power. Instead, in the integer case, the
# electric heater can only operate at either 0% or in-between 25% and 100% of its nominal power. For this reason, the
# final power profile on the city district level is not as "flat" as in the convex case. However, the integer case
# might better reflect the actual operation conditions of a real electric heater unit.


if __name__ == '__main__':
    # Run example:
    main(do_plot=True)



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import pyomo.environ as pyomo

from pycity_scheduling.classes.electrical_entity import ElectricalEntity


class ElectricalNode(ElectricalEntity):
    """
    Class ElectricalNode for scheduling purposes. Represents a building as a grid node inside an electrical grid within
    pycity_scheduling.

    Parameters
    ----------
    environment : pycity_scheduling.classes.Environment
        Common to all other objects. Includes time and weather instances.
    related_entity : Building, optional
        The entity coupled to the grid node.
    position : list, optional
         List of two floats specifying the position of the node inside the grid.
    min_V : float, optional
        Minimum voltage allowed for healthy operation in p.u.
    max_V : float, optional
        Maximum voltage allowed for healthy operation in p.u.
    ratio_pq_power : float, optional
        The fixed ratio of reactive power demand with respect to the active power demand, for the building connected
        to this node.
    """
    def __init__(self, environment, related_entity=None, position=None, min_V=0.95, max_V=1.05,
                 ratio_pq_power=1.0/3.0, *args, **kwargs):
        super().__init__(environment, *args, **kwargs)
        self._kind = "electricalnode"
        self._long_id = "EN_" + self._id_string
        self.related_entity = related_entity
        self.position = position
        self.min_V = min_V
        self.max_V = max_V
        self.ratio_pq_power = ratio_pq_power

        # Node power injection:
        self.new_var("p_node_el_vars")
        self.new_var("q_node_el_vars")

        # Square of node voltage:
        self.new_var("V_el_vars")

    def populate_model(self, model, mode='convex'):
        """
        Add electrical node block to pyomo ConcreteModel.

        Call parent's `populate_model` method and set variables lower bound to `None`.
        Set bound constraints on node voltage.
        
        Parameters
        ----------
        model : pyomo.ConcreteModel
        mode : str, optional
            Specifies which set of constraints to use.

            - `convex`  : Use linear constraints
            - `integer`  : Use same constraints as convex mode
        """
        super().populate_model(model, mode)
        m = self.model

        # Introduce power injection of the node as an additional variable for readability
        m.p_node_el_vars = pyomo.Var(m.t, domain=pyomo.Reals, bounds=(None, None), initialize=1)

        def p_node_el_vars_rule(model, t):
            return m.p_node_el_vars[t] == m.p_el_vars[t]
        m.p_node_el_vars_constr = pyomo.Constraint(m.t, rule=p_node_el_vars_rule)

        # Power injection is negative when the node is consuming power and positive when the node is injecting power:
        m.p_el_vars.setlb(None)
        m.q_el_vars = pyomo.Var(m.t, domain=pyomo.Reals)

        # Bounded voltage for healthy operation:
        m.V_el_vars = pyomo.Var(m.t, domain=pyomo.Reals, bounds=(self.min_V**2, self.max_V**2), initialize=1)



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import os.path as op
import calendar
import warnings
import pycity_base.classes.prices as pr


class Prices(pr.Prices):
    """
    Extension of pyCity_base class Prices for scheduling purposes.

    Parameters
    ----------
    timer : Timer
        Timer instance for generating needed prices.
    da_prices : array_like, optional
        Day-ahead prices for each timestep in the `simu_horizon`
        in [ct/kWh].
    tou_prices : array_like, optional
        Time-of-use prices for each timestep in the `simu_horizon`
        in [ct/kWh].
    co2_prices : array_like, optional
        CO2 emissions for each timestep in the `simu_horizon` in [g/kWh].
    feedin_factor : float, optional
        Factor which is multiplied to the prices for feed-in revenue.
        Should be in [0,1], as prices for feed-in are usually lower than
        for consumption.

    Notes
    -----
    - If prices are loaded automatically, the simulation period must lie within a single year.

    - CO2 emissions and day-ahead prices are currently available for the year 2015 only.
    """

    da_price_cache = None
    tou_price_cache = None
    tou_price_cache_year = None
    co2_price_cache = None

    def __init__(self, timer, da_prices=None,
                 tou_prices=None, co2_prices=None, feedin_factor=0):
        super(Prices, self).__init__()

        root_dir = op.dirname(op.dirname(__file__))
        timesteps = timer.time_in_year()
        factor = 900 / timer.time_discretization

        if da_prices is None:
            if self.da_price_cache is None:
                file_name = "da_prices_quarter-hour_2015.txt"
                file_path = op.join(root_dir, "data", file_name)
                tmp = np.loadtxt(file_path, dtype=np.float32, skiprows=1)
                tmp /= 10  # convert [Eur/MWh] to [ct/kWh]
                Prices.da_price_cache = tmp
            self.da_prices = self._interp_prices(Prices.da_price_cache,
                                                 timesteps,
                                                 timer.simu_horizon,
                                                 factor,
                                                 'step')
        else:
            if len(da_prices) != timer.simu_horizon:
                raise ValueError(
                    "Provided day-ahead price data do not match the number of "
                    "timesteps in the simulation horizon.\n"
                    "Number of timesteps: {0}, Length of price data: {1}"
                    .format(timer.simu_horizon, len(da_prices))
                )
            self.da_prices = np.array(da_prices)

        if tou_prices is None:
            if Prices.tou_price_cache_year != timer.year:
                summer_weekday_prices = np.array(
                    [7.7]*28 + [11.4]*16 + [14]*24 + [11.4]*8 + [7.7]*20
                )
                winter_weekday_prices = np.array(
                    [7.7]*28 + [14]*16 + [11.4]*24 + [14]*8 + [7.7]*20
                )
                weekend_prices = np.array([8.7] * 96)
                winter_week = np.concatenate((
                    np.tile(winter_weekday_prices, 5),
                    np.tile(weekend_prices, 2)
                ))
                summer_week = np.concatenate((
                    np.tile(summer_weekday_prices, 5),
                    np.tile(weekend_prices, 2)
                ))
                days_to_summer = 120 * 96
                if timer.is_leap:
                    days_to_summer += 96
                wd = calendar.weekday(timer.year, 1, 1) * 96
                winter_prices = np.tile(winter_week, 19)[wd:wd+days_to_summer]
                wd = calendar.weekday(timer.year, 5, 1) * 96
                summer_prices = np.tile(summer_week, 28)[wd:wd+184*96]
                wd = calendar.weekday(timer.year, 11, 1) * 96
                winter_prices2 = np.tile(winter_week, 15)[wd:wd+61*96]
                tou_prices = np.concatenate(
                    (winter_prices, summer_prices, winter_prices2)
                )
                tou_avg = 9.5
                file_name = "consumer_prices_yearly.txt"
                file_path = op.join(root_dir, "data", file_name)
                tmp = np.loadtxt(file_path, dtype=np.float32, skiprows=1)
                if len(tmp) <= timer.year - 2000:
                    warnings.warn("Year {} not in `consumer_prices_yearly.txt`. Using year {} instead."
                                  .format(timer.year, 2000+len(tmp)-1))
                    avg_price = tmp[-1]
                else:
                    avg_price = tmp[timer.year - 2000]
                Prices.tou_price_cache = tou_prices / tou_avg * avg_price
                Prices.tou_price_cache_year = timer.year
            self.tou_prices = self._interp_prices(Prices.tou_price_cache,
                                                  timesteps,
                                                  timer.simu_horizon,
                                                  factor,
                                                  'step')
        else:
            if len(tou_prices) != timer.simu_horizon:
                raise ValueError(
                    "Provided time of use prices do not match the number of "
                    "timesteps in the simulation horizon.\n"
                    "Number of timesteps: {0}, Length of price data: {1}"
                    .format(timer.simu_horizon, len(tou_prices))
                )
            self.tou_prices = np.array(tou_prices)

        if co2_prices is None:
            if self.co2_price_cache is None:
                file_name = "co2_emissions_quarter-hour_2015.txt"
                file_path = op.join(root_dir, "data", file_name)
                tmp = np.loadtxt(file_path, dtype=np.float32, skiprows=1)
                Prices.co2_price_cache = tmp
            self.co2_prices = self._interp_prices(Prices.co2_price_cache,
                                                  timesteps,
                                                  timer.simu_horizon,
                                                  factor,
                                                  'step')
        else:
            if len(co2_prices) != timer.simu_horizon:
                raise ValueError(
                    "Provided CO2 price data do not match the number of "
                    "timesteps in the simulation horizon.\n"
                    "Number of timesteps: {0}, Length of price data: {1}"
                    .format(timer.simu_horizon, len(co2_prices))
                )
            self.co2_prices = np.array(co2_prices)

        self.feedin_factor = feedin_factor

    @staticmethod
    def _interp_prices(prices, timesteps, length, factor=1, mode='linear'):
        """
        Interpolate and slice prices.

        Interpolates a price vector to the correct resolution and then extracts
        the specified range from it.

        Parameters
        ----------
        prices : array_like
            Original prices. Is not modified by the function.
        timesteps : int
            First timestep in the interpolated prices.
        length : int
            Number of timesteps in the interpolated prices.
        factor : float, optional
            Factor by which the original prices are interpolated.
        mode : str, optional
            Must be 'linear' or 'step'.
            'linear' : linear interpolation
            'step' : view original prices as step function (only possible if 'factor' is an integer value)

        Returns
        -------
        numpy.ndarray :
            Interpolated prices.
        """
        if factor != 1:
            old_len = len(prices)
            new_len = int(old_len * factor)
            if (factor > 1) and (factor != int(factor)):
                prices = np.interp(range(new_len), range(old_len), prices)
            elif mode == 'linear':
                prices = np.interp(range(new_len), range(old_len), prices)
            elif mode == 'step':
                if factor > 1:
                    factor = int(factor)
                    prices = np.array([np.repeat(p, factor) for p in prices])
                    np.reshape(prices, new_len)
                else:
                    factor = int(1/factor)
                    prices = prices[::factor].copy()
            else:
                raise ValueError("Mode {} is invalid.".format(mode))
        interp_prices = prices[timesteps:timesteps+length].copy()
        return interp_prices
